/*

cGMP_interface.h -- interface to the GMP float functions with caching  
                    for the iRRAM library
 
Copyright (C) 2004 Norbert Mueller/Daniel Meissner/Michael Bahra
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/
#ifndef cGMP_interface
#define cGMP_interface


#include "gmp.h"

#include "GMP_intrat.h"


#define MP_min    GMP_min
#define MP_max    GMP_max

/****** Type definitions ******/

#define MP_type	     hash_record*
#define MP_int_type  int_gmp_type
#define MP_rat_type  rat_gmp_type


/****** Initialization functions ******/

/* Backend initialization (if necessary) */
#define MP_initialize	{}
//initialization is handeld via a static variable hashtable_init_marker in cGMP_ext.cc
// because of static REAL variables, the initialization has to be done very early!


/* Initialization of MP/integer/rational variables */
#define MP_init(z)	z=NULL;
#define MP_int_init(z)	z=int_gmp_init()
#define MP_rat_init(z)	z=rat_gmp_init()

/* Deletion of MP/integer/rational variables */
#define MP_clear(z)	rfc_clear(&z)
#define MP_int_clear(z) int_gmp_free(z)
#define MP_rat_clear(z) rat_gmp_free(z)


/****** typechanging functions ******/
#define MP_mp_to_double(z)          mpf_get_d(z->data)
#define MP_double_to_mp(d,z)        hash_double_to_mp(d,&(z))

#define MP_int_to_mp(i,z)  hash_int_to_mp(i,&(z))

#define MP_int_to_INTEGER(i,z) 	int_gmp_int2integer(i,z)

#define MP_INTEGER_to_mp(i,z)   hash_INTEGER_to_mp(i,&(z))

#define MP_INTEGER_to_int(z) 	int_gmp_integer2int(z)

#define MP_mp_to_INTEGER(r,i)           mpz_set_f(i,r->data)

#define MP_double_to_INTEGER(i,z)   mpz_set_d(z,i)
#define MP_string_to_INTEGER(s,z,b) mpz_set_str(z,s,b)

#define MP_int_to_RATIONAL(i,z)          mpq_set_si(z,i,1)
#define MP_intint_to_RATIONAL(i,j,z)     {mpq_set_si(z,i,j);mpq_canonicalize(z);}
#define MP_double_to_RATIONAL(d,z)       mpq_set_d(z,d)
#define MP_string_to_RATIONAL(s,z)       rat_gmp_string_2_rat(z,s)
#define MP_INTEGER_to_RATIONAL(i,r)      mpq_set_z(r,i)
#define MP_INTINTEGER_to_RATIONAL(i,j,r) {mpq_set_num(r,i);mpq_set_den(r,j);mpq_canonicalize(r);}


/* duplicate value z1 to z2, with/without initialization of z2 */
#define MP_duplicate_w_init(z1,z2)           cGMP_d_w_i(z1,z2)
#define MP_duplicate_wo_init(z1,z2)          cGMP_d_wo_i(z1,z2)

#define MP_int_duplicate_w_init(z1,z2)       int_gmp_duplicate_w_init(z1,&(z2))
#define MP_int_duplicate_wo_init(z1,z2)      int_gmp_duplicate_wo_init(z1,z2) 

#define MP_rat_duplicate_w_init(z1,z2)       rat_gmp_duplicate_w_init(z1,&(z2))
#define MP_rat_duplicate_wo_init(z1,z2)      rat_gmp_duplicate_wo_init(z1,z2) 

/* copy z1 to z2, but precision p is sufficient */
#define MP_copy(z1,z2,p)                     cGMP_d_wo_i(z1,z2)

/****** simple arithmetic functions ******/

/* Multiple precision arithmetic, deterministic results */


/* Multiple precision arithmetic, deterministic results */
#define MP_add(z1,z2,z,p)   hash_add_det(z1,z2,&z,p)
#define MP_sub(z1,z2,z,p)   hash_sub_det(z1,z2,&z,p)
#define MP_mul(z1,z2,z,p)   hash_mul_det(z1,z2,&z,p)
#define MP_div(z1,z2,z,p)   hash_div_det(z1,z2,&z,p)
#define MP_addi(z1,z2,z,p)  hash_addi_det(z1,z2,&z,p)
#define MP_subi(z1,z2,z,p)  hash_subi_det(z1,z2,&z,p)
#define MP_isub(z1,z2,z,p)  hash_isub_det(z1,z2,&z,p)
#define MP_muli(z1,z2,z,p)  hash_muli_det(z1,z2,&z,p)
#define MP_divi(z1,z2,z,p)  hash_divi_det(z1,z2,&z,p)
#define MP_idiv(z1,z2,z,p)  hash_idiv_det(z1,z2,&z,p)

#define MP_abs(z1,z)        hash_abs(z1,&z)

/* Multiple precision arithmetic, multi-valued results
  The following versions of the arithmetic operations
  are allowed to be muli-valued, here in the sense that
  even for absolutely identical parameters there can be
  different results. Most backends simply use the 
  deterministic versions again...
*/    
#define MP_mv_add(z1,z2,z,p)  hash_add(z1,z2,&z,p)
#define MP_mv_mul(z1,z2,z,p)  hash_mul(z1,z2,&z,p)
#define MP_mv_sub(z1,z2,z,p)  hash_sub(z1,z2,&z,p)
#define MP_mv_div(z1,z2,z,p)  hash_div(z1,z2,&z,p)
#define MP_mv_idiv(z1,z2,z,p) hash_idiv(z1,z2,&z,p)
#define MP_mv_isub(z1,z2,z,p) hash_isub(z1,z2,&z,p)
#define MP_mv_divi(z1,z2,z,p) hash_divi(z1,z2,&z,p)
#define MP_mv_addi(z1,z2,z,p) hash_addi(z1,z2,&z,p)
#define MP_mv_subi(z1,z2,z,p) hash_subi(z1,z2,&z,p)
#define MP_mv_muli(z1,z2,z,p) hash_muli(z1,z2,&z,p)

/* INTEGER arithmetic, deterministic results */
#define MP_int_add(z1,z2,z)     int_gmp_add(z1,z2,z)
#define MP_int_sub(z1,z2,z)     int_gmp_sub(z1,z2,z)
#define MP_int_mul(z1,z2,z)     int_gmp_mul(z1,z2,z)
#define MP_int_div(z1,z2,z)     int_gmp_div(z1,z2,z)
#define MP_int_add_ui(z1,z2,z)  int_gmp_add_ui(z1,z2,z)
#define MP_int_sub_ui(z1,z2,z)  int_gmp_sub_ui(z1,z2,z)
#define MP_int_mul_si(z1,z2,z)  int_gmp_mul_si(z1,z2,z)
#define MP_int_div_ui(z1,z2,z)  int_gmp_div_ui(z1,z2,z)
#define MP_int_abs(z1,z)        int_gmp_abs(z1,z)
#define MP_int_neg(z1,z)        int_gmp_neg(z1,z)

/* RATIONAL arithmetic, deterministic results */
#define MP_rat_add(z1,z2,z)    rat_gmp_add(z1,z2,z)
#define MP_rat_sub(z1,z2,z)    rat_gmp_sub(z1,z2,z)
#define MP_rat_mul(z1,z2,z)    rat_gmp_mul(z1,z2,z)
#define MP_rat_div(z1,z2,z)    rat_gmp_div(z1,z2,z) 
#define MP_rat_add_ui(z1,z2,z) rat_gmp_add_ui(z1,z2,z)
#define MP_rat_sub_ui(z1,z2,z) rat_gmp_sub_ui(z1,z2,z)
#define MP_rat_mul_si(z1,z2,z) rat_gmp_mul_si(z1,z2,z)
#define MP_rat_div_si(z1,z2,z) rat_gmp_div_si(z1,z2,z)
#define MP_rat_si_div(z1,z2,z) rat_gmp_si_div(z1,z2,z)
#define MP_rat_abs(z1,z)       rat_gmp_abs(z1,z)

/* Additional integer/rational arithmetic */
#define MP_int_sqrt(z1,z)        int_gmp_sqrt(z1,z);
#define MP_int_root(z1,z2,z)     int_gmp_root(z1,z2,z);
#define MP_int_power_i(z1,z2,z)  int_gmp_power_i(z1,z2,z);
#define MP_int_power_ii(z1,z2,z) int_gmp_power_ii(z1,z2,z);
#define MP_int_fac(z1,z)         int_gmp_fac(z1,z);
#define MP_int_modulo(z1,z2,z)   int_gmp_modulo(z1,z2,z);
#define MP_int_log(z)            int_gmp_log(z);
#define MP_rat_power(z1,z2,z)    rat_gmp_power(z1,z2,z);
#define MP_rat_power2(z1,z2,z)   rat_gmp_power_ii(z1,z2,z);
#define MP_rat_get_numerator(z1,z) mpq_get_num(z,z1) 
#define MP_rat_get_denominator(z1,z) mpq_get_den(z,z1)
#define MP_rat_canon(z)          mpq_canonicalize(z) 

/* MP_compare: values -1 (lessthan), 0 (equal), or 1 (greater than) */
#define MP_compare(z1,z2)        mpf_cmp(z1->data,z2->data)
#define MP_int_compare(z1,z2)    int_gmp_cmp(z1,z2)
#define MP_rat_compare(z1,z2)    rat_gmp_cmp(z1,z2)

/* MP_sign: values -1 (negative), 0 (zero), or 1 (positive) */
#define MP_sign(z)          mpf_sgn(z->data)
#define MP_int_sign(z)      int_gmp_sgn(z)
#define MP_rat_sign(z)      rat_gmp_sgn(z)

/****** size information ******/
/* long MP_size : smallest(?) value k such that 2**k > |z|  */
#define MP_size(z)        ext_gmp_size(z->data)
#define MP_getsize(z,s)   ext_gmp_getsize(z->data,(ext_gmp_sizetype*)&s)
#define MP_int_size(z)    int_gmp_size(z)

/* truncate z1 yielding z2 */ 
#define MP_truncate(z1,z) hash_truncate(z1,&z);

/*Conversion to strings */
#define MP_swrite(z,w)   ext_gmp_swritee(z->data,w)

/* In/Out */
#define MP_writee(z,w)   ext_gmp_writee(z->data,w)
#define MP_writef(z,w,d) mpf_out_str(stdout,10,w,z->data)
#define MP_write(z,c)    ext_gmp_write(z->data,c)

#define MP_int_writee(z,w)  int_gmp_writee(z,w)
#define MP_int_swritee(z,w) int_gmp_swritee(z,w)
#define MP_int_writef(z,w)  int_gmp_outstr(z,w)
#define MP_int_write(z,c)   int_gmp_write(z,c)
#define MP_int_printf(z)    int_gmp_printf(z)
#define MP_int_sprintf(z)   int_gmp_sprintf(z)

#define MP_rat_writee(z,w)  rat_gmp_writee(z,w)
#define MP_rat_writef(z,w)  rat_gmp_outstr(z,w)
#define MP_rat_swritee(z,w) rat_gmp_swritee(z,w)
#define MP_rat_sprintf(z)   rat_gmp_sprintf(z)

/* Statistics, optional */

#define MP_var_count ext_gmp_var_count
//#define MP_space_count 
//#define MP_max_space_count

/* special functions, optional for the backends */
#define MP_sqrt(z1,z,p)      hash_sqrt_det(z1,&z,p)
#define MP_mv_sqrt(z1,z,p)   hash_sqrt(z1,&z,p)
#define MP_shift(z1,z,n)     hash_shift(z1,&z,n)
#define MP_int_shift(z1,z,n) int_gmp_shift(z1,z,n)
#define MP_rat_shift(z,z1,n) rat_gmp_shift(z1,z,n)


/***********************************************************/
/* Declaration of those types/functions that are necessary */
/* for user programs:                                       */ 
/***********************************************************/




typedef __mpf_struct  * ext_gmp_type;
typedef struct {unsigned int mantissa; int exponent; } ext_gmp_sizetype;
void ext_gmp_free(ext_gmp_type z);
void ext_gmp_getsize(ext_gmp_type z,ext_gmp_sizetype* s);
void ext_gmp_initialize(void);
ext_gmp_type ext_gmp_init(void);

extern int ext_gmp_var_count;
extern ext_gmp_type gmp_FreeVars[];
extern int gmp_FreeVarCount;
#define MaxFreeVars 1000

const unsigned int mp_bytes_per_limb=sizeof (mp_limb_t);
const int limb_size_in_bit=8* sizeof (mp_limb_t);
const int half_limb_size_in_bit=4* sizeof (mp_limb_t);
const mp_limb_t one_shift_limb_size_in_bit_1 = 1ul<<(8* sizeof (mp_limb_t)-1);
const mp_limb_t one_shift_half_limb_size_in_bit =1ul<<(4* sizeof (mp_limb_t));

inline void ext_gmp_initialize()
{
mpf_set_default_prec(53);
}


inline ext_gmp_type ext_gmp_init()
{
  ext_gmp_var_count +=1;
  if (gmp_FreeVarCount>0) {
    gmp_FreeVarCount -=1;
    return gmp_FreeVars[gmp_FreeVarCount];
  } else {
  ext_gmp_type z = (ext_gmp_type) malloc(sizeof (__mpf_struct));
  mpf_init(z);
  return z;
  }
}


inline void ext_gmp_free(ext_gmp_type z)
{ 
  if ( gmp_FreeVarCount < MaxFreeVars ) {
    gmp_FreeVars[gmp_FreeVarCount]=z;
    gmp_FreeVarCount +=1;
  } else {
    mpf_clear(z); 
    free (z);
  }
  ext_gmp_var_count -=1;
}

inline void ext_gmp_getsize(ext_gmp_type z,ext_gmp_sizetype* s)
{
  if ( z->_mp_size==0 ) {
    s->mantissa=0;
    s->exponent=GMP_min;
  } else if ( ((z->_mp_d)[abs(z->_mp_size)-1]==0) ) {
    s->mantissa=0;
    s->exponent=GMP_min;
  } else {
#if  __GMP_BITS_PER_MP_LIMB == 32
    s->mantissa=(z->_mp_d)[abs(z->_mp_size)-1];
    s->exponent=(z->_mp_exp-1) * limb_size_in_bit;

    if ( s->mantissa < one_shift_half_limb_size_in_bit )
      { s->mantissa = s->mantissa <<half_limb_size_in_bit;
        s->exponent-=half_limb_size_in_bit;
        if ( abs(z->_mp_size)>1 ) 
          s->mantissa+= (z->_mp_d)[abs(z->_mp_size)-2]
                         >>half_limb_size_in_bit;
    }
    if (s->mantissa > one_shift_limb_size_in_bit_1) {
      s->mantissa=(s->mantissa>>1)+1;
      s->exponent+=1;
    }
    else 
    if (abs(z->_mp_size)>0) s->mantissa+=1;
#else  
    long e;
    s->mantissa= ((unsigned int)(mpf_get_d_2exp(&e,z)*16777216.0)) + 1;
    s->exponent = e - 24;
#endif
  }
}

typedef struct {ext_gmp_type data;unsigned long long id;unsigned int rfc;int prec;} hash_record;
typedef hash_record * hash_record_pointer;

#define SAVE_RECORD_LIST_SIZE 16384
extern hash_record *save_record_list[SAVE_RECORD_LIST_SIZE];
extern unsigned int save_record_list_pointer;

unsigned int hashtable_init();
hash_record_pointer getrecord();


inline hash_record_pointer getrecord(){ 
  hash_record* r;
  if (save_record_list_pointer==0){
     r = new(hash_record);
     r->data=ext_gmp_init();
  } else {
     r = save_record_list[--save_record_list_pointer];
  }
  r->rfc=0;
  return r;
  };
  

inline void freerecord(hash_record* r){ 
  if (save_record_list_pointer<SAVE_RECORD_LIST_SIZE  ){
    save_record_list[save_record_list_pointer++]=r;
  }
  else{
    ext_gmp_free(r->data);
    delete(r);
  }
};

inline void rfc_clear(hash_record_pointer * z){
    if((*z)!=NULL &&--((*z)->rfc) == 0){
       freerecord(*z);
    }
} 

inline void cGMP_d_w_i (MP_type z1,MP_type& z2){z2=z1;z2->rfc++;}
inline void cGMP_d_wo_i(MP_type z1,MP_type& z2){rfc_clear(&(z2));z2=z1;z1->rfc++;}

#endif /*ifndef cGMP_interface*/
