/*

iRRAM.h -- central header file for the iRRAM library
 
Copyright (C) 2001-2003 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/


/*
Authors:  all by Norbert, except: 

  2003-07 INTERVAL extensions partly by Shao Qi

  2001-07 INTEGER and RATIONAL partly by Tom van Diessen
  

*/

#ifndef iRRAM_H
#define iRRAM_H

#include <string>
using namespace std;


// forward declaration of some classes

class INTEGER;
class RATIONAL;
class DYADIC;
class LAZY_BOOLEAN;
class REAL;
class COMPLEX;

#if defined(LRGMP)
	#include "LRGMP_interface.h"
	#define iRRAM_BACKEND "LRGMP"
#elif defined(cGMP)
	#include "cGMP_interface.h"
	#define iRRAM_BACKEND "cGMP"
#elif defined(GMP)
	#include "GMP_interface.h"
	#define iRRAM_BACKEND "GMP"
#elif defined(MPFR)
	#include "MPFR_interface.h"
	#define iRRAM_BACKEND "MPFR"
#else
	#include "Backend not defined"
#endif


extern "C" void iRRAM_initialize(int argc,char** argv);

template <class ARGUMENT, class RESULT> 
 RESULT iRRAM_exec (RESULT f(ARGUMENT), ARGUMENT);

#ifndef do_not_activate_iRRAM_control
extern void compute(void); 

int iRRAM_compute(int arg){compute();return 0;};
 
int main (int argc,char **argv){
  iRRAM_initialize(argc,argv);
  return iRRAM_exec(iRRAM_compute,0);
}
#endif



extern char* iRRAM_error_msg[];
class iRRAM_Numerical_Exception
{
public:
iRRAM_Numerical_Exception(const int msg){type=msg;};
//private:
int type;
};
#define ERRORDEFINE(x,y) x,
enum iRRAM_execption_list {
#include "iRRAM_errno.h"
};



#define SIZETYPEMANTISSA unsigned int
#define SIZETYPEEXPONENT int

struct sizetype { SIZETYPEMANTISSA  mantissa; SIZETYPEEXPONENT exponent; } ;



#include "iRRAM_LAZYBOOLEAN.h"
#include "iRRAM_REAL.h"
#include "iRRAM_DYADIC.h"
#include "iRRAM_INTEGER.h"
#include "iRRAM_RATIONAL.h"
#include "iRRAM_REALMATRIX.h"
#include "iRRAM_SPARSEREALMATRIX.h"
#include "iRRAM_COMPLEX.h"
#include "iRRAM_INTERVAL.h"
#include "iRRAM_STREAMS.h"



#define ABSOLUTE 0
#define RELATIVE 1
void precision_policy (int policy);
void stiff_begin ();
void stiff_end   ();
void continous_begin ();
void continous_end   ();





REAL atoREAL(const char* s);


 

/****************************************************************************/
// arithmetic functions
/****************************************************************************/
REAL power(const REAL& x, const REAL& y);
REAL power(const REAL& x, int n);
REAL modulo  (const REAL& x, const REAL& y);
REAL maximum (const REAL& x, const REAL& y);
REAL minimum (const REAL& x, const REAL& y);

/****************************************************************************/
// roots
/****************************************************************************/
REAL sqrt    (const REAL& x);
REAL root    (const REAL& x,int n);

/****************************************************************************/
// trigonometric functions
/****************************************************************************/
REAL sin     (const REAL& x);
REAL cos     (const REAL& x);
REAL tan     (const REAL& x);
REAL cotan   (const REAL& x);
REAL sec     (const REAL& x);
REAL cosec   (const REAL& x);

/****************************************************************************/
// inverse trigonometric functions
/****************************************************************************/
REAL atan    (const REAL& x);
REAL asin    (const REAL& x);
REAL acos    (const REAL& x);
REAL acotan  (const REAL& x);
REAL asec    (const REAL& x);
REAL acosec  (const REAL& x);

/****************************************************************************/
//hyperbolic functions
/****************************************************************************/
REAL sinh    (const REAL& x);
REAL cosh    (const REAL& x);
REAL tanh    (const REAL& x);
REAL coth    (const REAL& x);
REAL sech    (const REAL& x);
REAL cosech  (const REAL& x);

/****************************************************************************/
// inverse hyperbolic functions 
/****************************************************************************/
REAL asinh   (const REAL& x);
REAL acosh   (const REAL& x);
REAL atanh   (const REAL& x);
REAL acoth   (const REAL& x);
REAL asech   (const REAL& x);
REAL acosech (const REAL& x);

/****************************************************************************/
// exponentiation + logarithm
/****************************************************************************/
REAL exp     (const REAL& x);
REAL log     (const REAL& x);

/****************************************************************************/
// special constants values                                                 
/****************************************************************************/
REAL pi      ();   // = 3.141592653...
REAL euler   ();   // = 2.718281828...
REAL ln2     ();   // = 0.693147180...

/****************************************************************************/
//  matrix functions
/****************************************************************************/
REALMATRIX exp    (const REALMATRIX& x);

REALMATRIX steady_state (const REALMATRIX& x);

REALMATRIX steady_state (const SPARSEREALMATRIX& x);
REALMATRIX equilib (const SPARSEREALMATRIX& x);
void equilib_del (SPARSEREALMATRIX& x,REALMATRIX& z);




/****************************************************************************/
// templates for limit operators
/****************************************************************************/

template <class ARGUMENT, class RESULT> 
RESULT limit_mv (RESULT f(int prec,
                                         int* choice,
                                         const ARGUMENT&),
                           const ARGUMENT& x);

template <class ARGUMENT, class RESULT> 
RESULT limit (RESULT f(int prec,const ARGUMENT&),
                           const ARGUMENT& x);


template <class ARGUMENT, class DISCRETE, class RESULT> 
RESULT limit (RESULT f(int prec,const ARGUMENT&, DISCRETE),
                           const ARGUMENT& x, DISCRETE param);





#endif
/* ! iRRAM_H */
