/****************************************************************/
/* Definition of class RATIONAL                                 */
/****************************************************************/

class RATIONAL
{
public:

friend class INTEGER;
friend class DYADIC;
friend class REAL;

/****** Constructors ******/

RATIONAL(MP_rat_type y);

RATIONAL();
RATIONAL(const int i);
RATIONAL(const double d);
RATIONAL(const char* s);
RATIONAL(const INTEGER& y);
RATIONAL(const DYADIC& y); 
RATIONAL(const RATIONAL& y); 

RATIONAL(const int i, const int j);
RATIONAL(const INTEGER& x, const INTEGER& y);

/****** Copy constructor ******/

RATIONAL& operator = (const RATIONAL& y);
RATIONAL& operator = (const INTEGER& y);
RATIONAL& operator = (const int y);

/****** Destructor ******/

~RATIONAL();

/****** Standard arithmetic ******/

friend RATIONAL  operator +  (const RATIONAL& x, const RATIONAL& y);
friend RATIONAL  operator +  (const RATIONAL& x, const int	 y);
friend RATIONAL  operator +  (const RATIONAL& x, const INTEGER&  y);
friend RATIONAL  operator +  (const int       x, const RATIONAL& y);
friend RATIONAL  operator +  (const INTEGER&  x, const RATIONAL& y);

friend RATIONAL& operator += (      RATIONAL& x, const RATIONAL& y);
friend RATIONAL& operator += (      RATIONAL& x, const INTEGER&  y);
friend RATIONAL& operator += (      RATIONAL& x, const int	 y);


friend RATIONAL  operator -  (const RATIONAL& x, const RATIONAL& y);
friend RATIONAL  operator -  (const RATIONAL& x, const int	 y);
friend RATIONAL  operator -  (const RATIONAL& x, const INTEGER&  y);
friend RATIONAL  operator -  (const int       x, const RATIONAL& y);
friend RATIONAL  operator -  (const INTEGER&  x, const RATIONAL& y);

friend RATIONAL& operator -= (      RATIONAL& x, const RATIONAL& y);
friend RATIONAL& operator -= (      RATIONAL& x, const INTEGER&  y);
friend RATIONAL& operator -= (      RATIONAL& x, const int	 y);

friend RATIONAL  operator -  (const RATIONAL& x);

friend RATIONAL  operator *  (const RATIONAL& x, const RATIONAL& y);
friend RATIONAL  operator *  (const RATIONAL& x, const int	 y);
friend RATIONAL  operator *  (const RATIONAL& x, const INTEGER&  y);
friend RATIONAL  operator *  (const int       x, const RATIONAL& y);
friend RATIONAL  operator *  (const INTEGER&  x, const RATIONAL& y);

friend RATIONAL& operator *= (      RATIONAL& x, const RATIONAL& y);
friend RATIONAL& operator *= (      RATIONAL& x, const INTEGER&  y);
friend RATIONAL& operator *= (      RATIONAL& x, const int	 y);

friend RATIONAL  operator /  (const RATIONAL& x, const RATIONAL& y);
friend RATIONAL  operator /  (const RATIONAL& x, const int	 y);
friend RATIONAL  operator /  (const RATIONAL& x, const INTEGER&  y);
friend RATIONAL  operator /  (const int       x, const RATIONAL& y);
friend RATIONAL  operator /  (const INTEGER&  x, const RATIONAL& y);

friend RATIONAL& operator /= (      RATIONAL& x, const RATIONAL& y);
friend RATIONAL& operator /= (      RATIONAL& x, const INTEGER&  y);
friend RATIONAL& operator /= (      RATIONAL& x, const int	 y);

friend RATIONAL	 scale  	 (const RATIONAL& x, const int k);

friend RATIONAL  sqrt		 (const RATIONAL& x);
friend RATIONAL	 root		 (const RATIONAL& x, const int n);
friend RATIONAL	 abs		 (const RATIONAL& x);
friend RATIONAL	 sqrt_int	 (const RATIONAL& x);

friend INTEGER  numerator	 (const RATIONAL& x);
friend INTEGER  denominator	 (const RATIONAL& x);
friend int 	sign	 	 (const RATIONAL& x);

/****** Comparisons ******/

friend bool 	operator <      (const RATIONAL& x, const RATIONAL& y);
friend bool 	operator <=     (const RATIONAL& x, const RATIONAL& y);
friend bool 	operator >      (const RATIONAL& x, const RATIONAL& y);
friend bool 	operator >=     (const RATIONAL& x, const RATIONAL& y);
friend bool 	operator ==     (const RATIONAL& x, const RATIONAL& y);
friend bool 	operator !=     (const RATIONAL& x, const RATIONAL& y);

/****** String conversion ******/

friend string    swrite          (const RATIONAL& x);
friend string    swrite          (const RATIONAL& x, const int w);

/****** Private ******/
private:

MP_rat_type value;
RATIONAL(MP_int_type y);
};
