/*

iRRAM_limit_templates.h -- template definitions file for the limit operators of the iRRAM library
 
Copyright (C) 2005 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/
#ifndef iRRAM_limit_templates_h
#define iRRAM_limit_templates_h

#include "iRRAM_core.h"

template <class ARGUMENT, class RESULT>
RESULT limit (RESULT f(int prec,const ARGUMENT&),
                           const ARGUMENT& x)

{
  ITERATION_STACK SAVED_STACK;
  ACTUAL_STACK.inlimit+=1;

  RESULT lim,limnew;
  sizetype limnew_error,element_error;
  sizetype lim_error,x_error;
  int element=ACTUAL_STACK.actual_prec;
  int firsttime=2;

  x.geterror(x_error);

  ACTUAL_STACK.prec_inc=int(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
  ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc;

  DEBUG1(2,"starting limit_gen1...\n");
  while (1) {
    try {
    DEBUG2(2,"trying to compute limit_gen1 with precicion 2^(%d)...\n",element);
    limnew=f(element,x);
    sizetype_set(element_error,1,element);
    limnew.geterror(limnew_error);
    sizetype_inc(limnew_error,element_error);
    if (firsttime ==2 ) if ( limnew_error.exponent > SAVED_STACK.actual_prec -SAVED_STACK.prec_inc/ACTUAL_STACK.prec_factor
    	&&  limnew_error.exponent > x_error.exponent -SAVED_STACK.prec_inc/ACTUAL_STACK.prec_factor) {
    DEBUG0(2,{cerr<<"computation not precise enough ("
                  << limnew_error.mantissa <<"*2^"<< limnew_error.exponent
                  <<"), trying normal p-sequence\n";});
       element=4+iRRAM_starting_prec;
       ACTUAL_STACK.prec_inc=0;
       firsttime=1;
    }
    if ( firsttime != 0 || sizetype_less(limnew_error,lim_error) ) {
      lim=limnew;
      lim_error=limnew_error;
      DEBUG3(2,"getting result with error %d*2^(%d)\n",
               lim_error.mantissa, lim_error.exponent);
      } else {
      DEBUG1(2,"computation successful, but no improvement\n");
      }
    firsttime=0;
    if (element<=SAVED_STACK.actual_prec)break;
    ACTUAL_STACK.prec_inc=int(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
    element=iRRAM_starting_prec+ACTUAL_STACK.prec_inc;
    }
    catch ( Iteration it) {
      if ( firsttime==0) {
      DEBUG1(2,"computation failed, using best success\n");
      break;
      } else
      if ( firsttime==2) {
      DEBUG1(2,"computation failed, trying normal p-sequence\n");
      element=4+iRRAM_starting_prec;
      ACTUAL_STACK.prec_inc=0;
      firsttime=1;
      } else {
      DEBUG1(1,"computation of limit_gen1 failed totally\n");
      REITERATE(0);
      }}
  }
  lim.seterror(lim_error);
  DEBUG0(2,{cerr<<"end of limit_gen1 with error "
                << lim_error.mantissa <<"*2^("<< lim_error.exponent<<")\n";});
  return lim;
}

template <class ARGUMENT, class DISCRETE, class RESULT>
RESULT limit (RESULT f(int prec,const ARGUMENT&,DISCRETE param),
                           const ARGUMENT& x,DISCRETE param)

{
  ITERATION_STACK SAVED_STACK;
  ACTUAL_STACK.inlimit+=1;

  RESULT lim,limnew;
  sizetype limnew_error,element_error;
  sizetype lim_error,x_error;
  int element=ACTUAL_STACK.actual_prec;
  int firsttime=2;

  ACTUAL_STACK.prec_inc=int(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
  ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc;

  x.geterror(x_error);
  DEBUG1(2,"starting limit_gen1...\n");
  while (1) {
   try {
    DEBUG2(2,"trying to compute limit_gen1 with precicion 2^(%d)...\n",element);
    limnew=f(element,x,param);
    sizetype_set(element_error,1,element);
    limnew.geterror(limnew_error);
    sizetype_inc(limnew_error,element_error);
    if (firsttime ==2 ) if ( limnew_error.exponent > SAVED_STACK.actual_prec -SAVED_STACK.prec_inc/ACTUAL_STACK.prec_factor
                       &&  limnew_error.exponent > x_error.exponent -SAVED_STACK.prec_inc/ACTUAL_STACK.prec_factor) {
    DEBUG0(2,{cerr<<"computation not precise enough ("
                  << limnew_error.mantissa <<"*2^"<< limnew_error.exponent
                  <<"), trying normal p-sequence\n";});
       element=4+iRRAM_starting_prec;
       ACTUAL_STACK.prec_inc=0;
       firsttime=1;
    }
    if ( firsttime  != 0 || sizetype_less(limnew_error,lim_error) ) {
      lim=limnew;
      lim_error=limnew_error;
      DEBUG3(2,"getting result with error %d*2^(%d)\n",
               lim_error.mantissa, lim_error.exponent);
      } else {
      DEBUG1(2,"computation successful, but no improvement\n");
      }
    firsttime=0;
    if (element<=SAVED_STACK.actual_prec)break;
    ACTUAL_STACK.prec_inc=int(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
    element=iRRAM_starting_prec+ACTUAL_STACK.prec_inc;
    }
    catch ( Iteration it) {
      if ( firsttime==0) {
      DEBUG1(2,"computation failed, using best success\n");
      break;
      } else
      if ( firsttime==2) {
      DEBUG1(2,"computation failed, trying normal p-sequence\n");
      element=4+iRRAM_starting_prec;
      ACTUAL_STACK.prec_inc=0;
      firsttime=1;
      } else {
      DEBUG1(1,"computation of limit_gen1 failed totally\n");
      REITERATE(0);
      }}

  }
  lim.seterror(lim_error);
  DEBUG0(2,{fprintf(stderr,"end of limit_gen1 with error %d*2^(%d)\n",
                   lim_error.mantissa,lim_error.exponent);});
  return lim;
}

template <class ARGUMENT, class RESULT>
RESULT limit_mv (RESULT f(int prec,
                          int* choice,
                          const ARGUMENT&),
                          const ARGUMENT& x)
{
  ITERATION_STACK SAVED_STACK;
  ACTUAL_STACK.inlimit+=1;

  RESULT lim,limnew;
  sizetype limnew_error,element_error;
  int choice=0;
  sizetype lim_error,x_error;
  int element=ACTUAL_STACK.actual_prec;
  int firsttime=2;

  if ( (ACTUAL_STACK.inlimit==0) && !cache_i.get(choice)) cache_i.put(choice);


  ACTUAL_STACK.prec_inc=int(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
  ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc;

  x.geterror(x_error);
  DEBUG1(2,"starting limit_mv...\n");
  while (1) {
   try {
    DEBUG2(2,"trying to compute limit_mv with precicion 2^(%d)...\n",element);
    limnew=f(element,&choice,x);
    if ( SAVED_STACK.inlimit==0 ) cache_i.modify(choice);
    sizetype_set(element_error,1,element);
    limnew.geterror(limnew_error);
    sizetype_inc(limnew_error,element_error);
    if (firsttime ==2 ) if ( limnew_error.exponent > SAVED_STACK.actual_prec -SAVED_STACK.prec_inc/ACTUAL_STACK.prec_factor &&  limnew_error.exponent > x_error.exponent -SAVED_STACK.prec_inc/ACTUAL_STACK.prec_factor) {
    DEBUG0(2,{fprintf(stderr,"computation not precise enough (%d*2^%d), trying normal p-sequence\n",
                   limnew_error.mantissa,limnew_error.exponent);});
       element=4+iRRAM_starting_prec;
       ACTUAL_STACK.prec_inc=0;
       firsttime=1;
    }}
     catch ( Iteration it) {
      if ( firsttime==0) {
      DEBUG1(2,"computation failed, using best success\n");
      break;
      } else
      if ( firsttime==2) {
      DEBUG1(2,"computation failed, trying normal p-sequence\n");
      element=4+iRRAM_starting_prec;
      ACTUAL_STACK.prec_inc=0;
      firsttime=1;
      } else {
      DEBUG1(1,"computation of limit_gen1 failed totally\n");
      REITERATE(0);
      }}
   if ( firsttime != 0 || sizetype_less(limnew_error,lim_error) ) {
      lim=limnew;
      lim_error=limnew_error;
      DEBUG3(2,"getting result with error %d*2^(%d)\n",
               lim_error.mantissa, lim_error.exponent);
      } else {
      DEBUG1(2,"computation successful, but no improvement\n");
      }
    firsttime=0;
    if (element<=SAVED_STACK.actual_prec)break;
    ACTUAL_STACK.prec_inc=int(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
    element=iRRAM_starting_prec+ACTUAL_STACK.prec_inc;
    }
  lim.seterror(lim_error);
  DEBUG0(2,{fprintf(stderr,"end of limit_mv with error %d*2^(%d)\n",
                   lim_error.mantissa,lim_error.exponent);});
  return lim;
}

template <class ARGUMENT, class DISCRETE, class RESULT>
RESULT  limit_lip (RESULT  f(int,const ARGUMENT&,DISCRETE param),
            int lip_value,
	    bool on_domain(const ARGUMENT&),
            const ARGUMENT& x,DISCRETE param)
{
  if ( on_domain(x) != true ) REITERATE(0);

  ITERATION_STACK SAVED_STACK;
  ACTUAL_STACK.inlimit+=1;

  ACTUAL_STACK.prec_inc=ACTUAL_STACK.prec_inc+iRRAM_prec_inc2;
  ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc;

  ARGUMENT x_new;
  RESULT lim;
  sizetype lim_error,xerror;

  x_new=x;
  sizetype_exact(xerror);
  x_new.seterror(xerror);
  x.geterror(xerror);

  DEBUG1(2,"starting limit_lip ...\n");
  while (1) {
     try{
      DEBUG2(2,"trying to compute limit_lip with precision %d...\n",ACTUAL_STACK.actual_prec);
    lim=f(SAVED_STACK.actual_prec,x_new,param);
    lim.geterror(lim_error);
    if (lim_error.exponent > SAVED_STACK.actual_prec ) {
      ACTUAL_STACK.prec_inc=int(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
      ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc;
      DEBUG2(2,"limit_lip too imprecise, increasing precision locally to %d...\n",ACTUAL_STACK.actual_prec);
    } else {
      DEBUG3(2,"getting result with local error %d*2^(%d)\n",
             lim_error.mantissa, lim_error.exponent);
      break;
    }}
    catch ( Iteration it){
      ACTUAL_STACK.prec_inc=int(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
      ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc;
      DEBUG2(2,"limit_lip failed, increasing precision locally to %d...\n",ACTUAL_STACK.actual_prec);
    } }
  sizetype_set(lim_error,1,SAVED_STACK.actual_prec);
  lim.adderror(lim_error);
  sizetype_shift(lim_error,xerror,lip_value);
  lim.adderror(lim_error);
  DEBUG0(2,{lim.geterror(lim_error);
            fprintf(stderr,"end of limit_lip with error %d*2^(%d)\n",
              lim_error.mantissa,lim_error.exponent);
            x.geterror(lim_error);
            fprintf(stderr,"  error of argument: %d*2^(%d)\n",
              lim_error.mantissa,lim_error.exponent);});
  return lim;
}


#endif 
/* ! iRRAM_limit_templates_h */
