#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifndef DEFINITION_Dyadic_Base
#include "LRGMP_Dyadic_Base.h"
#endif

#ifndef DEFINITION_Dyadic
#include "LRGMP_Dyadic.h"
#endif

#include "Dyadic_Base.c"  

#include "gmp.h"
#include "gmp-impl.h"

#define   max(a,b)        (a < b ? b : a)
#define   min(a,b)        (a < b ? a : b)
#define   FALSE	0
#define   TRUE	1

Dyadic_LRErrorProc Dyadic_LRError;
int Dyadic_LRVarCount;



#define BLKFACTOR	10240
/* BLKFACTOR: Number of blocks allocated at once if list of free blocks
   is empty, slightly reduces memory consumption.
*/

typedef Dyadic_Base_WertFeld WertFeld;

static WertFeld AKKUMULATOR1;
static WertFeld AKKUMULATOR2;
static WertFeld AKKUMULATOR3;
static double basis;
static double lnbasis;
static int LRSpaceLow;
static int LRmaxSpaceLow;
static Dyadic_Dyadic CONV_HILFE, LRINT_ARG, A_LRSQRT, B_LRSQRT;
static int trigger;
static void Initialize (void);
inline void LRSET (Dyadic_Dyadic *Z, int HighExp, Dyadic_Base_WertFeld AKKU, int Anzahl, BOOLEAN NEGATIV);
inline void LRGET (const Dyadic_Dyadic *Z, int HighExp, Dyadic_Base_WertFeld AKKU, int Anzahl);
static void LRTODEC (const Dyadic_Dyadic *Z, int A, char Dez[], unsigned int , int *DExp);

inline BlockZeiger LRNewBlock(void);
inline void LRFreeBlock(BlockZeiger q);
void Dyadic_LRABSVAL (Dyadic_Dyadic *Z);
void Dyadic_LRSPLIT (const Dyadic_Dyadic *X, Dyadic_Dyadic *Y, Dyadic_Dyadic *Z);

static BlockZeiger LRFreeBlocks=0L;
static BlockZeiger LRDummyBlock;

static Dyadic_Dyadic LRFreeVars=0L;

inline BlockZeiger LRNewBlock(void)
{
    BlockZeiger p;
    LRSpaceLow++;
    LRmaxSpaceLow=max(LRmaxSpaceLow,LRSpaceLow);
    if (! LRFreeBlocks)  {
      int i;
      BlockZeiger a;
      a = malloc( BLKFACTOR *sizeof(Block));
      LRFreeBlocks = a;
      for (i=1; i< BLKFACTOR; i++) {
        a->BlockNext = a+1;
        a++;
        }
      a->BlockNext=0L;
      }
    p= LRFreeBlocks;
    LRFreeBlocks=p->BlockNext;
    return p;
}

inline void LRFreeBlock(BlockZeiger q)
{
    q->BlockNext =LRFreeBlocks;
    LRFreeBlocks=q;
    LRSpaceLow--;
}

static void Initialize()
{
  int i;

  basis = 1.0;
  for (i = 1; i <= Dyadic_Base_BITZAHL; i += 1) {
    basis = 2.0 * basis;
  }
  lnbasis = log(basis);
  LRDummyBlock=LRNewBlock();
  Dyadic_LRError = Dyadic_LRErrorStatus;
  Dyadic_LRINIT(&CONV_HILFE);
  Dyadic_LRINIT(&LRINT_ARG);
  Dyadic_LRINIT(&A_LRSQRT);
  Dyadic_LRINIT(&B_LRSQRT);
  Dyadic_LRVarCount = 0;
  LRSpaceLow = 0;
  LRmaxSpaceLow = 0;
  trigger = 100000;
/*  printf("trigger fuer Karatsuba :\n");
  scanf("%d",&trigger);
  printf("\n");*/
}

void Dyadic_LRErrorIgnore (int Error)
{
}

void Dyadic_LRErrorStatus (int Error)
{
  fprintf(stderr,"Dyadic-Error, ");
  switch (Error) {
  case 1:;
    fprintf(stderr,"LRTOINT-Overflow\n");
    break;
  case 2:;
    fprintf(stderr,"Division durch 0\n");
    break;
  case 3:;
    fprintf(stderr,"Falscher CHAR bei DECTOLR\n");
    break;
  case 4:;
    fprintf(stderr,"Negatives Argument bei LRSQRT\n");
    break;
  default :
    fprintf(stderr,"Fehler %d gefunden\n",Error);
    break;
  }
}

void Dyadic_LRErrorHalt (int Error)
{
  Dyadic_LRErrorStatus(Error);
  exit(1);
}

int Dyadic_LRmaxSpace ()
{
  return LRmaxSpaceLow / 1024 * sizeof(Block);
}

int Dyadic_LRSpace ()
{
  return LRSpaceLow / 1024 * sizeof(Block);
}

unsigned int Dyadic_LOG2 (unsigned int N)
{
  unsigned int L;

  L = 0;
  for (;;) {
    if (N < 2) {
      goto EXIT_1;
    }
    L++;
    N = N / 2;
  } EXIT_1:;
  return L;
}

void Dyadic_LRCLEAR (Dyadic_Dyadic *Z)
{

  BlockZeiger p, q;

// printf("CLEAR: %lu\n",Z);

  p = ((Dyadic)(*Z))->Wert;
  while (p != 0L) {
    q = p;
    p = p->BlockNext;
    LRFreeBlock(q);
  }
  (*Z)->Wert=(void*)LRFreeVars;
  LRFreeVars=(*Z);
  (*Z)=0L;
  Dyadic_LRVarCount -= 1;
}

void Dyadic_LRINIT (Dyadic_Dyadic *Z)
{
  if (LRFreeVars) {
    (*Z) =LRFreeVars;
    LRFreeVars=(void*)(LRFreeVars->Wert);
  } else {
  (*Z)=malloc(sizeof(struct S_3));
  }
  Dyadic_LRVarCount++;
  ((Dyadic)(*Z))->Wert = 0L;
// printf("INIT:  %lu\n",Z);
}

void Dyadic_LRNEGATE (Dyadic_Dyadic *Z)
{
  ((Dyadic)(*Z))->negativ = !((Dyadic)(*Z))->negativ;
}

void Dyadic_LRABSVAL (Dyadic_Dyadic *Z)
{
  ((Dyadic)(*Z))->negativ = FALSE;
}

inline void LRSET
(Dyadic_Dyadic *Z, int HighExp, Dyadic_Base_WertFeld AKKU, int Anzahl, BOOLEAN NEGATIV)
{
  int MinAkku, MaxAkku;
  BlockZeiger p, q;
  mp_limb_t *MaxP,*MinP,*WertP;
/*
 {int ExpC;
 printf("LRSET: ");
 for (ExpC = 0;ExpC < Anzahl ; ExpC += 1) printf("%lu ",AKKU[ExpC]);
 printf("\n");
 }
*/

  MaxP = &AKKU[Anzahl - 1];
  MaxAkku = Anzahl;
  while (*MaxP == 0 && MaxP > AKKU) {
    MaxP--;
    MaxAkku--;
  }
  MinP = AKKU;
  MinAkku = 1;
  while (*MinP  == 0 && MaxP > MinP) {
    MinP++;
    MinAkku++;
  }
  if (((Dyadic)(*Z))->Wert == 0L) {
    p=LRNewBlock();
    ((Dyadic)(*Z))->Wert = p;
    p->BlockNext = 0L;
  } else {
    p = ((Dyadic)(*Z))->Wert;
  }



  while (MinP <= MaxP - BLOCKSIZE) {
     MaxP -= BLOCKSIZE;
     memcpy(p->BlockFeld,MaxP+1,sizeof(mp_limb_t)*BLOCKSIZE);
     if (p->BlockNext == 0L) {
          p->BlockNext=LRNewBlock();
          p = p->BlockNext;
          p->BlockNext = 0L;
        } else {
          p = p->BlockNext;
        }
    }



  WertP= &(p->BlockFeld[BLOCKSIZE-1]);

  while (MinP <= MaxP )  *(WertP--) = *(MaxP--);

  q = p->BlockNext;
  p->BlockNext = 0L;
  while (q != 0L) {
    p = q;
    q = q->BlockNext;
    LRFreeBlock(p);
  }
  {
    register struct S_3 *W_1 = (Dyadic)(*Z);

    W_1->anzahl = MaxAkku - MinAkku + 1;
    if (W_1->Wert->BlockFeld[BLOCKSIZE-1] == 0)
       W_1->highexponent=Dyadic_Base_MinInt;
    else
       W_1->highexponent = HighExp  - Anzahl + MaxAkku;
    W_1->negativ = NEGATIV;
  }
}

inline void LRGET
(const Dyadic_Dyadic *Z, int HighExp, Dyadic_Base_WertFeld AKKU, int Anzahl)
{
  int IndAkku, MinAkku, MaxAkku;
  mp_limb_t *AP,*ZP;
  BlockZeiger p;

   if (Anzahl >=  (Dyadic_Base_MAXSIZE>>1) ) {
      fprintf(stderr,"Akkumulator-Groesse ueberschritten!\n");
      exit(1);
    }
  {
    register struct S_3 *W_2 = (Dyadic)(*Z);

    if (W_2->highexponent < HighExp - Anzahl) {
      MinAkku = Anzahl;
      MaxAkku = Anzahl;
    } else {
      MinAkku = HighExp - W_2->highexponent;
      if (Anzahl > MinAkku + W_2->anzahl) {
        MaxAkku = MinAkku + W_2->anzahl;
      } else {
        MaxAkku = Anzahl;
      }
    }

    AP = &AKKU[Anzahl-1];
    p = W_2->Wert;


    IndAkku=1;
    while ( IndAkku++ <= MinAkku ) *(AP--) = 0;
    IndAkku=MinAkku+1;
 
    AP += 1;
    while ( IndAkku <= MaxAkku-BLOCKSIZE ){
      IndAkku+=BLOCKSIZE;
       AP -= BLOCKSIZE;
       memcpy(AP,p->BlockFeld,sizeof(mp_limb_t)*BLOCKSIZE);
       p = p->BlockNext;
    }

    AP = &AKKU[Anzahl-IndAkku];
    ZP = &p->BlockFeld[BLOCKSIZE-1];

    while ( IndAkku++ <= MaxAkku )  *(AP--) = *(ZP--);
    IndAkku=MaxAkku+1;
    while ( IndAkku++ <= Anzahl )   *(AP--) = 0;
  }

/*
 {int ExpC;
 printf("LRGET: ");
 for (ExpC = 0;ExpC < Anzahl ; ExpC += 1) printf("%lu ",AKKU[ExpC]);
 printf("\n");
 }
*/

}

void Dyadic_LRTOgmpint(const Dyadic_Dyadic *r, int_gmp_type i){
  unsigned int e = ((Dyadic)(*r))->highexponent;
  mp_ptr wp;

  LRGET(r, e, AKKUMULATOR1, e);
  if (i->_mp_alloc < e) _mpz_realloc (i, e);
 
  wp = i->_mp_d;
 
  MPN_COPY (wp, AKKUMULATOR1, e);
  if ( ((Dyadic)(*r))->negativ ) i->_mp_size = -e;
  else  i->_mp_size = e;
};

extern void Dyadic_gmpintTOLR(int_gmp_type i,Dyadic_Dyadic *r){
  if (mpz_size(i) == 0 ) {
    AKKUMULATOR1[0]=0;
    LRSET(r, 1, AKKUMULATOR1, 1, 0);
  } else {
    LRSET(r, mpz_size (i), i->_mp_d, mpz_size(i), mpz_sgn(i)<0);
  }
};



void Dyadic_INTTOLR
(int W, Dyadic_Dyadic *Z)
{
  long long M=W;
  if (W < 0) M = -M;
  AKKUMULATOR1[0] = M;
  LRSET(Z, 1, AKKUMULATOR1, 1, W < 0);
}

void Dyadic_LRWRITE
(const Dyadic_Dyadic *Z, int F)
{
  unsigned char  BIT [Dyadic_Base_BITZAHL];
  int i, j, n;
  unsigned int l;
  BlockZeiger p;
  int length;

  printf("HighExponent: %5d",((Dyadic)(*Z))->highexponent);
  printf(" Anzahl: %5d",((Dyadic)(*Z))->anzahl);
  printf(" Vorzeichen: ");
  if (((Dyadic)(*Z))->negativ) {
    printf("negativ");
  } else {
    printf("positiv");
  }
  printf("\n");
  p = ((Dyadic)(*Z))->Wert;
  n = BLOCKSIZE + 1;
  length = 64;
  for (i = ((Dyadic)(*Z))->anzahl; i >= 1; i += -1) {
    n--;
    if (n < 1) {
      p = p->BlockNext;
      n = BLOCKSIZE;
    }
    l = p->BlockFeld[n - 1];
    for (j = 1; j <= Dyadic_Base_BITZAHL; j += 1) {
      if (l % 2 != 0) {
        BIT[j - 1] = '1';
      } else {
        BIT[j - 1] = '0';
      }
      l = l / 2;
    }
    for (j = Dyadic_Base_BITZAHL; j >= 1; j += -1) {
      printf("%c",BIT[j - 1]);
      F++;
      if (F == 0) {
        printf("\n");
        return;
      }
      length--;
      if (length == 0) {
        printf("\n");
        length = 64;
      }
    }
  }
  for (i = F; i <= 0; i += 1) {
    printf("0");
    length--;
    if (length == 0) {
      printf("\n");
      length = 64;
    }
  }
  if (length != 64) {
    printf("\n");
  }
}

double Dyadic_LRTODOUBLE
(const Dyadic_Dyadic *X)
{
  double R;
  int exponent=((Dyadic)(*X))->highexponent*Dyadic_Base_BITZAHL;
  LRGET(X, ((Dyadic)(*X))->highexponent, AKKUMULATOR1, 3);
  R = ((AKKUMULATOR1[0]/basis + AKKUMULATOR1[1] )/basis + AKKUMULATOR1[2] )/  basis;
  if (((Dyadic)(*X))->negativ) R = -R;
  return ldexp(R,exponent);
}

void Dyadic_DOUBLETOLR(double X, Dyadic_Dyadic *Z)
{
  double H,R;
  int e,n,sgn=0;
  if (X<0){ sgn=1; X=-X;}
  R=frexp(X,&e);
  if (e > 0) {
    n = (e - 1) / Dyadic_Base_BITZAHL;
  } else {
    n = e / Dyadic_Base_BITZAHL;
    n--;
  }
  e=e-n*Dyadic_Base_BITZAHL;
  R=ldexp(R,e);
  R=modf(R,&H);
  AKKUMULATOR1[2]=H;
  R=ldexp(R,Dyadic_Base_BITZAHL);
  R=modf(R,&H);
  AKKUMULATOR1[1]=H;
  R=ldexp(R,Dyadic_Base_BITZAHL);
  R=modf(R,&H);
  AKKUMULATOR1[0]=H;
  LRSET( Z, n+1, AKKUMULATOR1, 3, sgn );
}

char* Dyadic_LRSWRITEE(const Dyadic_Dyadic *Z, int w)
{ int e;
  char * Dez;
  int e_abs;
  if (w < 9) {
    Dez=strndup("*********",w);
  } else {
    Dez=malloc(1+w);
    LRTODEC(Z, w - 6, Dez, w, &e);
    Dez[w-6]=0;
    e_abs=e; if (e_abs<0) e_abs=-e_abs;
    if (e_abs<10000){e_abs=4;}
    else if (e_abs<100000){e_abs=5;}
    else if (e_abs<1000000){e_abs=6;}
    else if (e_abs<10000000){e_abs=7;}
    else if (e_abs<100000000){e_abs=8;}
    else if (e_abs<1000000000){e_abs=9;}
    else {e_abs=10;};
    if (e_abs > w-2){sprintf(&Dez[w-6],"E*****");
    } else {
      if (e_abs==4){sprintf(&Dez[w-6],"E%+05d",e);}
      else if (e_abs==5){sprintf(&Dez[w-7],"E%+06d",e);}
      else if (e_abs==6){sprintf(&Dez[w-8],"E%+07d",e);}
      else if (e_abs==7){sprintf(&Dez[w-9],"E%+08d",e);}
      else if (e_abs==8){sprintf(&Dez[w-10],"E%+09d",e);}
      else if (e_abs==9){sprintf(&Dez[w-11],"E%+010d",e);}
      else {sprintf(&Dez[w-12],"E%+011d",e);};
    }
  }
 return Dez;
}


void Dyadic_LRTOINT
(const Dyadic_Dyadic *X, int *I)
{
  unsigned int e;

  e = ((Dyadic)(*X))->highexponent;
  LRGET(X, e, AKKUMULATOR1, 1);
  if (AKKUMULATOR1[0] == 0) {
    *I = 0;
  } else if (e > 1) {
    *I = 0;
    (*Dyadic_LRError)(1);
  } else if (e < 1) {
    *I = 0;
  } else {
    *I = AKKUMULATOR1[0];
  }
  if (((Dyadic)(*X))->negativ) {
    *I = -*I;
  }
}

static void LRTODEC
(const Dyadic_Dyadic *Z, int Size, char Dez[], unsigned int O_1, int *DExp)
{
  int R, L, i, j, n, m, h;
  unsigned int k;
  unsigned int NULLWERT;
  int Groesse;
  unsigned int * Dezi;

  Dezi=malloc(O_1*sizeof(unsigned int));



  Groesse = O_1-2;
  if (Groesse < Size) {
    L = Groesse;
  } else {
    L = Size;
  }

  NULLWERT = (int)('0');
  if (((Dyadic)(*Z))->highexponent > 0) {
    n = ((Dyadic)(*Z))->highexponent;
    m = n;
    LRGET(Z, m, AKKUMULATOR1, m);
    *DExp = 0;
    while (m > 0) {
      AKKUMULATOR3[0] = 10;
      Dyadic_Base_SysDiv(AKKUMULATOR1, AKKUMULATOR3, AKKUMULATOR2, m, 1, m, TRUE);
      Dezi[*DExp % Groesse] = AKKUMULATOR1[0];
      (*DExp)++;
      if (AKKUMULATOR2[m-1] == 0)  m--;
      for (j = 0;j<m; j += 1)  AKKUMULATOR1[j] = AKKUMULATOR2[j];  
    }
    if (*DExp <= L) {
      R = L - *DExp;
      m = 10 * R / (3 * Dyadic_Base_BITZAHL) + 3;
      LRGET(Z, n, AKKUMULATOR1, n + m);
      AKKUMULATOR3[0] = 1000000000;
      h=AKKUMULATOR3[0];
      for (i = 1; i<= R+1 ; i += 1) {
         if ( h==AKKUMULATOR3[0] ) Dyadic_Base_SysMul(AKKUMULATOR1, AKKUMULATOR3, AKKUMULATOR2, m, 1, m + 1);
         h=h/10;
         Dezi[Groesse - i] = AKKUMULATOR2[m]/h;
         AKKUMULATOR2[m]=AKKUMULATOR2[m]%h;
         if (h==1) {h=AKKUMULATOR3[0]; for (j = 1;j<m; j += 1)  AKKUMULATOR1[j] = AKKUMULATOR2[j];}
      }
    }
  } else {
    n = ((Dyadic)(*Z))->highexponent;
    *DExp = 0;
    m = 10 * L / (3 * Dyadic_Base_BITZAHL) + 3;
    LRGET(Z, n, AKKUMULATOR1, m);
    AKKUMULATOR3[0] = 10;
    for (;;) {
      Dyadic_Base_SysMul(AKKUMULATOR1, AKKUMULATOR3, AKKUMULATOR2, m, 1, m + 1);
      if (AKKUMULATOR2[0] == 0) {
        for (j = 0;j<=m; j += 1)  AKKUMULATOR2[j] = AKKUMULATOR2[j+1];
        m--;
	}
      if (m == 0) break;
      if (AKKUMULATOR2[m] != 0) {
        if (n == 0) break;
        for (j = 0;j<=m; j += 1)   AKKUMULATOR1[j] = AKKUMULATOR2[j];
        n++;
        m++;
      } else {
        for (j = 0; j<=m-1; j += 1)  AKKUMULATOR1[j] = AKKUMULATOR2[j];
      }
     (*DExp)--;
    }
    k = (Groesse + (*DExp - 1) % Groesse) % Groesse;
    Dezi[k] = AKKUMULATOR2[m];
    AKKUMULATOR3[0] = 1000000000;
    for (j = 0;j<m; j += 1) AKKUMULATOR1[j] = AKKUMULATOR2[j];
    h=AKKUMULATOR3[0];
    for (i = 1; i<= L; i += 1) {
          if ( h==AKKUMULATOR3[0] ) Dyadic_Base_SysMul(AKKUMULATOR1, AKKUMULATOR3, AKKUMULATOR2, m, 1, m+1);
          h=h/10;
          k = (Groesse + k - 1) % Groesse;
          Dezi[k] = AKKUMULATOR2[m]/h;
          AKKUMULATOR2[m]=AKKUMULATOR2[m]%h;
          if (h==1) {h=AKKUMULATOR3[0]; for (j = 0;j<m; j += 1)    AKKUMULATOR1[j] = AKKUMULATOR2[j];}
    }
  }
  if (((Dyadic)(*Z))->negativ) {
    Dez[0] = '-';
  } else {
    Dez[0] = '+';
  }
  Dez[1] = '.';
  i = 1;
  {
      for (j = *DExp - 1;; j += -1) {
        i++;
        Dez[i] = (unsigned char)(NULLWERT + Dezi[(Groesse + j % Groesse) % Groesse]);
        if (j <= *DExp - L) break;
      }
  }
  if (Dez[L] >= (unsigned char)(NULLWERT + 6) ) {
     for (i=L-1;i>1;i--) {
       if (Dez[i] != '9' ) {
         Dez[i]=Dez[i]+1;
         break;
       } else Dez[i]='0';
     }
     if (i==1) {Dez[2]='1';(*DExp)++;};
  }
//  Dez[1] = Dez[2];
//  Dez[2]='.';
//  (*DExp)--;
  Dez[L] = (unsigned char)(0);
  free(Dezi);
}

void Dyadic_DECTOLR
(const unsigned char Dez[], unsigned int O_2, int DExp, Dyadic_Dyadic *Z)
{
  unsigned int i;
  int j;

  Dyadic_INTTOLR(0, Z);
  i = 0;
  for (;;) {
    if (i == (O_2 - 1)) {
      goto EXIT_3;
    }
    if (Dez[i] == (unsigned char)(0)) {
      goto EXIT_3;
    }
    if (Dez[i] < '0' || Dez[i] > '9') {
      (*Dyadic_LRError)(3);
      j = 0;
    } else {
      j = (int)(Dez[i]) - (int)('0');
    }
    Dyadic_INTTOLR(j, &CONV_HILFE);
    Dyadic_LRMULTINT(Z, 10, Z, -2);
    Dyadic_LRADD(Z, &CONV_HILFE, Z, -2);
    i++;
  } EXIT_3:;
  if (DExp < 0) {
    Dyadic_INTTOLR(10, &CONV_HILFE);
    {
      int B_31 = 2, B_32 = -DExp;

      if (B_31 <= B_32)
        for (j = B_31;; j += 1) {
          Dyadic_LRMULTINT(&CONV_HILFE, 10, &CONV_HILFE, -2);
          if (j >= B_32) break;
        }
    }
    Dyadic_LRDIV(Z, &CONV_HILFE, Z, -Dyadic_LRSIZE(&CONV_HILFE));
  } else {
    {
      int B_33 = 1, B_34 = DExp;

      if (B_33 <= B_34)
        for (j = B_33;; j += 1) {
          Dyadic_LRMULTINT(Z, 10, Z, -2);
          if (j >= B_34) break;
        }
    }
  }
}

void Dyadic_LRWRITEE
(const Dyadic_Dyadic *Z, int L)
{
  int DExp;
  char * Dez;
  Dez=malloc(L*sizeof(unsigned char));

  if (L < 9) {
    printf("*********");
  } else {
    LRTODEC(Z, L - 6, Dez, L, &DExp);
    Dez[L-6]=0;
    printf("%s",Dez);
    printf("E");
    if (DExp < 0) {
      printf("-");
      DExp = -DExp;
    } else {
      printf("+");
    }
    printf("%04d",DExp);
  }
  free(Dez);
}

void Dyadic_LRWRITEF
(const Dyadic_Dyadic *Z, int L, int R)
{
  int DExp, n, i, j, m;
  char * Dez;

  Dyadic_INTTOLR(5, &CONV_HILFE);
  

  if (0 <= R)
    for (i = 0; i<= R ; i += 1) {
      Dyadic_LRDIVINT(&CONV_HILFE, 10, &CONV_HILFE, -R * 4);
     }

  if (Dyadic_LRSIGN(Z) == -1) {
    Dyadic_LRSUB(Z, &CONV_HILFE, &CONV_HILFE, -R * 4);
  } else {
    Dyadic_LRADD(Z, &CONV_HILFE, &CONV_HILFE, -R * 4);
  }
  Dez=malloc(L*sizeof(unsigned char));
  LRTODEC(&CONV_HILFE, L, Dez, L, &DExp);
  if (L - R - 2 >= DExp) {
    if (DExp <= 0) {
      n = 0;
      m = -DExp;
    } else {
      n = DExp;
      m = 0;
    }

    if (n + 1 <= L - R - 2)
      for (i = n + 1 ;; i += 1) {
          printf(" ");
          if (i >= L - R - 2) break;
        }

    printf("%c",Dez[0]);

    for (i = 1 ; i <= n ; i += 1) {
          printf("%c",Dez[i + 1]);
        }
    printf(".");
    if (m > R) {
      m = R;
    }
    for (i = 1 ; i <= m ; i += 1)  printf("0");

    {
      int B_47 = n + 1, B_48 = n + R - m;

      if (B_47 <= B_48)
        for (i = B_47;; i += 1) {
          printf("%c",Dez[i + 1]);
          if (i >= B_48) break;
        }
    }
  } else {
    if (L < 9) {
      {
        int B_49 = 1, B_50 = L;

        if (B_49 <= B_50)
          for (j = B_49;; j += 1) {
            printf("*");
            if (j >= B_50) break;
          }
      }
    } else {
      {
        int B_51 = 1, B_52 = L - 6;

        if (B_51 <= B_52)
          for (j = B_51;; j += 1) {
            printf("%c",Dez[j - 1]);
            if (j >= B_52) break;
          }
      }
      printf("E");
      if (DExp < 0) {
        printf("-");
        DExp = -DExp;
      } else {
        printf("+");
      }
      j = 1000;
      for (i = 1; i <= 4; i += 1) {
        printf("%1d",DExp / j);
        DExp = DExp % j;
        j = j / 10;
      }
    }
  }
  free(Dez);
}

void Dyadic_LRADD
(const Dyadic_Dyadic *X, const Dyadic_Dyadic *Y, Dyadic_Dyadic *Z, int N)
{
  int Anzahl, HighExp;
  int n;
  BOOLEAN AkkuNegativ;

  if (N > 0) {
    n = (N - 1) / Dyadic_Base_BITZAHL;
  } else {
    n = N / Dyadic_Base_BITZAHL;
    n--;
  }
  if (((Dyadic)(*X))->highexponent > ((Dyadic)(*Y))->highexponent) {
    HighExp = ((Dyadic)(*X))->highexponent;
  } else {
    HighExp = ((Dyadic)(*Y))->highexponent;
  }
  if (n > HighExp) {
    Dyadic_INTTOLR(0, Z);
    return;
  }
  if (((Dyadic)(*X))->highexponent - ((Dyadic)(*X))->anzahl > ((Dyadic)(*Y))->highexponent - ((Dyadic)(*Y))->anzahl) {
    Anzahl = HighExp - ((Dyadic)(*Y))->highexponent + ((Dyadic)(*Y))->anzahl;
  } else {
    Anzahl = HighExp - ((Dyadic)(*X))->highexponent + ((Dyadic)(*X))->anzahl;
  }
  if (Anzahl > HighExp - n) {
    Anzahl = HighExp - n;
  }
  if ( Anzahl > 0) {
  LRGET(X, HighExp, AKKUMULATOR1, Anzahl);
  LRGET(Y, HighExp, AKKUMULATOR2, Anzahl);
  if (((Dyadic)(*X))->negativ == ((Dyadic)(*Y))->negativ) {
    Dyadic_Base_SysAdd(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, Anzahl);
    LRSET(Z, HighExp + 1, AKKUMULATOR3, Anzahl + 1, ((Dyadic)(*X))->negativ);
  } else {
    Dyadic_Base_SysSub(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, Anzahl, &AkkuNegativ);
    if (AkkuNegativ) {
      LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, !((Dyadic)(*X))->negativ);
    } else {
      LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, ((Dyadic)(*X))->negativ);
    }
  }
  } else {
  AKKUMULATOR3[0]=0;
  LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, 0);
  }
}

void Dyadic_LRSUB
(const Dyadic_Dyadic *X, const Dyadic_Dyadic *Y, Dyadic_Dyadic *Z, int N)
{
  int Anzahl, HighExp;
  int n;
  BOOLEAN AkkuNegativ;

  if (N > 0) {
    n = (N - 1) / Dyadic_Base_BITZAHL;
  } else {
    n = N / Dyadic_Base_BITZAHL;
    n--;
  }
  if (((Dyadic)(*X))->highexponent > ((Dyadic)(*Y))->highexponent) {
    HighExp = ((Dyadic)(*X))->highexponent;
  } else {
    HighExp = ((Dyadic)(*Y))->highexponent;
  }
  if (n > HighExp) {
    Dyadic_INTTOLR(0, Z);
    return;
  }
  if (((Dyadic)(*X))->highexponent - ((Dyadic)(*X))->anzahl > ((Dyadic)(*Y))->highexponent - ((Dyadic)(*Y))->anzahl) {
    Anzahl = HighExp - ((Dyadic)(*Y))->highexponent + ((Dyadic)(*Y))->anzahl;
  } else {
    Anzahl = HighExp - ((Dyadic)(*X))->highexponent + ((Dyadic)(*X))->anzahl;
  }
  if (Anzahl > HighExp - n) {
    Anzahl = HighExp - n;
  }
  if ( Anzahl > 0) {
  LRGET(X, HighExp, AKKUMULATOR1, Anzahl);
  LRGET(Y, HighExp, AKKUMULATOR2, Anzahl);
  if (((Dyadic)(*X))->negativ != ((Dyadic)(*Y))->negativ) {
    Dyadic_Base_SysAdd(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, Anzahl);
    LRSET(Z, HighExp + 1, AKKUMULATOR3, Anzahl + 1, ((Dyadic)(*X))->negativ);
  } else {
    Dyadic_Base_SysSub(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, Anzahl, &AkkuNegativ);
    if (AkkuNegativ) {
      LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, !((Dyadic)(*X))->negativ);
    } else {
      LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, ((Dyadic)(*X))->negativ);
    }
  }
  } else {
  AKKUMULATOR3[0]=0;
  LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, 0);
  }
}

void Dyadic_LRMULT
(const Dyadic_Dyadic *X, const Dyadic_Dyadic *Y, Dyadic_Dyadic *Z, int N)
{
  int HighExp, Anzahl, XAnz, YAnz;
  int n;

  if (N - 2 >= 0) {
    n = (N - 2) / Dyadic_Base_BITZAHL;
  } else {
    n = (N - 1) / Dyadic_Base_BITZAHL;
    n--;
  }
  HighExp = ((Dyadic)(*X))->highexponent + ((Dyadic)(*Y))->highexponent;
  if (n >= HighExp) {
    Dyadic_INTTOLR(0, Z);
    return;
  }
  Anzahl = HighExp - n;
  XAnz = ((Dyadic)(*X))->anzahl;
  if (XAnz > Anzahl) {
    XAnz = Anzahl;
  }
  LRGET(X, ((Dyadic)(*X))->highexponent, AKKUMULATOR1, XAnz);
  YAnz = ((Dyadic)(*Y))->anzahl;
  if (YAnz > Anzahl) {
    YAnz = Anzahl;
  }
  LRGET(Y, ((Dyadic)(*Y))->highexponent, AKKUMULATOR2, YAnz);
  if (XAnz > 0 && YAnz > 0 && Anzahl > 0) {
    if ( Anzahl > XAnz+YAnz ) Anzahl=XAnz+YAnz;
    Dyadic_Base_SysMul(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, XAnz, YAnz, Anzahl);
    LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, ((Dyadic)(*X))->negativ != ((Dyadic)(*Y))->negativ);
  } else {
    AKKUMULATOR3[0]=0;
    LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, 0);
  }
}


void Dyadic_LRDIV
(const Dyadic_Dyadic *X, const Dyadic_Dyadic *Y, Dyadic_Dyadic *Z, int N)
{
  int YAnzahl, Anzahl;
  int n;

  if (Dyadic_LRSIGN(Y) == 0) {
    Dyadic_INTTOLR(0, Z);
    (*Dyadic_LRError)(2);
    return;
  }
  if (N - 2 > 0) {
    n = (N - 2) / Dyadic_Base_BITZAHL;
  } else {
    n = (N - 1) / Dyadic_Base_BITZAHL;
    n--;
  }
  Anzahl = ((Dyadic)(*X))->highexponent - ((Dyadic)(*Y))->highexponent - n + 1;
  if (Anzahl <= 0) {
    Dyadic_INTTOLR(0, Z);
  } else {
  
    if (((Dyadic)(*Y))->anzahl < Anzahl + 1) {
      YAnzahl = ((Dyadic)(*Y))->anzahl;
    } else {
      YAnzahl = Anzahl + 1;
    }
 
    LRGET(X, ((Dyadic)(*X))->highexponent, AKKUMULATOR1, Anzahl);
    LRGET(Y, ((Dyadic)(*Y))->highexponent, AKKUMULATOR2, YAnzahl);
    Dyadic_Base_SysDiv(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, Anzahl, YAnzahl, Anzahl, FALSE);
    LRSET(Z, 1 + ((Dyadic)(*X))->highexponent - ((Dyadic)(*Y))->highexponent, AKKUMULATOR3, Anzahl, ((Dyadic)(*X))->negativ != ((Dyadic)(*Y))->negativ);
  }
}

void Dyadic_LRSHIFT
(const Dyadic_Dyadic *X, Dyadic_Dyadic *Z, int I)
{
  int n;
  unsigned int ShiftZahl;
  unsigned int Anzahl;

  Anzahl = ((Dyadic)(*X))->anzahl + 1;
  LRGET(X, ((Dyadic)(*X))->highexponent + 1, AKKUMULATOR1, Anzahl);
  if (I >= 0) {
    n = I / Dyadic_Base_BITZAHL;
  } else {
    n = (I + 1) / Dyadic_Base_BITZAHL;
    n--;
  }
  ShiftZahl = I - n * Dyadic_Base_BITZAHL;
  if (ShiftZahl != 0) {
    Dyadic_Base_SysBlockShl(AKKUMULATOR1, Anzahl, ShiftZahl);
  }
  LRSET(Z, ((Dyadic)(*X))->highexponent + 1 + n, AKKUMULATOR1, Anzahl, ((Dyadic)(*X))->negativ);
}

void Dyadic_LRSQUARE
(const Dyadic_Dyadic *X, Dyadic_Dyadic *Z, int N)
{
  int HighExp, Anzahl, XAnz;
  int n;

  if (N >= 2) {
    n = (N - 2) / Dyadic_Base_BITZAHL;
  } else {
    n = (N - 1) / Dyadic_Base_BITZAHL;
    n--;
  }
  HighExp = ((Dyadic)(*X))->highexponent + ((Dyadic)(*X))->highexponent;
  if (n >= HighExp) {
    Dyadic_INTTOLR(0, Z);
    return;
  }
  Anzahl = HighExp - n;
  XAnz = ((Dyadic)(*X))->anzahl;
  if (XAnz > Anzahl) {
    XAnz = Anzahl;
  }
  LRGET(X, ((Dyadic)(*X))->highexponent, AKKUMULATOR1, XAnz);
  if ( Anzahl > 2*XAnz ) Anzahl=2*XAnz;
  Dyadic_Base_SysMul(AKKUMULATOR1, AKKUMULATOR1, AKKUMULATOR3, XAnz, XAnz, Anzahl);
  LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, FALSE);
}

void Dyadic_LRADDINT
(const Dyadic_Dyadic *X, int Y, Dyadic_Dyadic *Z, int N)
{
  Dyadic_INTTOLR(Y, &LRINT_ARG);
  Dyadic_LRADD(X, &LRINT_ARG, Z, N);
}

void Dyadic_LRSUBINT
(const Dyadic_Dyadic *X, int Y, Dyadic_Dyadic *Z, int N)
{
  Dyadic_INTTOLR(Y, &LRINT_ARG);
  Dyadic_LRSUB(X, &LRINT_ARG, Z, N);
}

void Dyadic_LRINTSUB
(int Y, const Dyadic_Dyadic *X, Dyadic_Dyadic *Z, int N)
{
  Dyadic_INTTOLR(Y, &LRINT_ARG);
  Dyadic_LRSUB(&LRINT_ARG, X, Z, N);
}

void Dyadic_LRINTDIV
(int Y, const Dyadic_Dyadic *X, Dyadic_Dyadic *Z, int N)
{
  Dyadic_INTTOLR(Y, &LRINT_ARG);
  Dyadic_LRDIV(&LRINT_ARG, X, Z, N);
}


void Dyadic_LRMULTINT
(const Dyadic_Dyadic *X, int Y, Dyadic_Dyadic *Z, int N)
{
  Dyadic_INTTOLR(Y, &LRINT_ARG);
  Dyadic_LRMULT(X, &LRINT_ARG, Z, N);
}

void Dyadic_LRDIVINT
(const Dyadic_Dyadic *X, int Y, Dyadic_Dyadic *Z, int N)
{
  Dyadic_INTTOLR(Y, &LRINT_ARG);
  Dyadic_LRDIV(X, &LRINT_ARG, Z, N);
}

int Dyadic_LRSIGN
(const Dyadic_Dyadic *X)
{
  if (((Dyadic)(*X))->Wert->BlockFeld[BLOCKSIZE - 1] == 0) {
    return 0;
  } else {
    if (((Dyadic)(*X))->negativ) {
      return -1;
    } else {
      return 1;
    }
  }
}

void Dyadic_LRGETSIZE
(const Dyadic_Dyadic *X,LRsizetype* s)
{
  if (((Dyadic)(*X))->Wert->BlockFeld[BLOCKSIZE - 1]  == 0) {
    s->exponent=Dyadic_Base_MinInt;
    s->mantissa=0;
  } else {
#if Dyadic_Base_BITZAHL == 32
    s->exponent=Dyadic_Base_BITZAHL * (((Dyadic)(*X))->highexponent - 1);
    if ( s->mantissa <1<<(Dyadic_Base_BITZAHL/2) )
      { s->mantissa = s->mantissa <<(Dyadic_Base_BITZAHL/2);
        s->exponent-=Dyadic_Base_BITZAHL/2;
        if ( ((Dyadic)(*X))->anzahl > 1) 
          s->mantissa+= ((((Dyadic)(*X))->Wert->BlockFeld[BLOCKSIZE - 2])
                         >>(Dyadic_Base_BITZAHL-Dyadic_Base_BITZAHL/2));
    }
    if (s->mantissa > 1<<(Dyadic_Base_BITZAHL-1)) {
        s->mantissa=(s->mantissa>>1)+1;
        s->exponent+=1;
    }
    else
    if ( ((Dyadic)(*X))->anzahl > 1) s->mantissa +=1;
  }
#else
    int e;
    double z=((double)((Dyadic)(*X))->Wert->BlockFeld[BLOCKSIZE - 1]);
    if ( ((Dyadic)(*X))->anzahl > 1) z=z +ldexp((double)((Dyadic)(*X))->Wert->BlockFeld[BLOCKSIZE - 2],-Dyadic_Base_BITZAHL);
    z=frexp(z,&e);
    s->mantissa= ((unsigned int)(z*16777216.0)) + 1;
    s->exponent =Dyadic_Base_BITZAHL * (((Dyadic)(*X))->highexponent - 1) +  e - 24;
//fprintf(stderr,"LRGETSIZE: %lu,%d\n", s->mantissa, s->exponent) ;
#endif
  }
}



int Dyadic_LRSIZE
(const Dyadic_Dyadic *X)
{
  mp_limb_t c, r, m,n;
  c = ((Dyadic)(*X))->Wert->BlockFeld[BLOCKSIZE - 1];

  if (c == 0) {
    return Dyadic_Base_MinInt;
  } else {
    n=c;
    r=0; 
    for (m=Dyadic_Base_BITZAHL>>1; m>0; m=m>>1 ) {
      if (n>= (((mp_limb_t) 1)<<m) ){
        r+=m; n=n>>m;
      }
    }
    if (c >(((mp_limb_t) 1)<<r) || ((Dyadic)(*X))->anzahl >1) r++;
//fprintf(stderr,"LRSIZE: %g\n",r + Dyadic_Base_BITZAHL * (((Dyadic)(*X))->highexponent - 1));
    return r + Dyadic_Base_BITZAHL * (((Dyadic)(*X))->highexponent - 1);
  }
}

int Dyadic_LRCOMPARE
(const Dyadic_Dyadic *X, const Dyadic_Dyadic *Y)
{ int sx,sy,lx,ly,erg,i,anz;
  sx=Dyadic_LRSIGN(X);
  sy=Dyadic_LRSIGN(Y);
  if (sx < sy) return -1;
  if (sx > sy) return +1;
  if (sx == 0)  return  0;
  if (sx < 0)  erg=-1; else erg=+1;
  lx=((Dyadic)(*X))->highexponent;
  ly=((Dyadic)(*X))->highexponent;
  if (lx<ly) return erg;
  if (ly>lx) return -erg;
  LRGET(X, lx, AKKUMULATOR1, 4);
  LRGET(Y, ly, AKKUMULATOR2, 4);
  for (i=3;i>=0;i--) {
    if (AKKUMULATOR1[i]<AKKUMULATOR2[i]) return erg;
    if (AKKUMULATOR1[i]>AKKUMULATOR2[i]) return -erg;
  }
  anz=max (((Dyadic)(*X))->anzahl,((Dyadic)(*Y))->anzahl);
  LRGET(X, lx, AKKUMULATOR1, anz);
  LRGET(Y, ly, AKKUMULATOR2, anz);
  for (i=anz-1;i>=4;i--) {
    if (AKKUMULATOR1[i]<AKKUMULATOR2[i]) return erg;
    if (AKKUMULATOR1[i]>AKKUMULATOR2[i]) return -erg;
  }
  return 0;
}

int Dyadic_LRLENGTH
(const Dyadic_Dyadic *X)
{
  return ((Dyadic)(*X))->anzahl * Dyadic_Base_BITZAHL;
}

void Dyadic_LRCOPY
(const Dyadic_Dyadic *X, Dyadic_Dyadic *Z, int N)
{
  int n;

  if (N > 0) {
    n = (N - 1) / Dyadic_Base_BITZAHL;
  } else {
    n = N / Dyadic_Base_BITZAHL;
    n--;
  }
  if (n >= ((Dyadic)(*X))->highexponent) {
    Dyadic_INTTOLR(0, Z);
  } else {
    LRGET(X, ((Dyadic)(*X))->highexponent, AKKUMULATOR1, ((Dyadic)(*X))->highexponent - n + 1);
    LRSET(Z, ((Dyadic)(*X))->highexponent, AKKUMULATOR1, ((Dyadic)(*X))->highexponent - n + 1, ((Dyadic)(*X))->negativ);
  }
}
void Dyadic_LRABS
(const Dyadic_Dyadic *X, Dyadic_Dyadic *Z)
{
    Dyadic_LRDUPLICATE(X,Z);
    Dyadic_LRABSVAL(Z);
}

void Dyadic_LRDUPLICATE
(const Dyadic_Dyadic *X, Dyadic_Dyadic *Z)
{
    BlockZeiger px=((Dyadic)(*X))->Wert;
    BlockZeiger pz,qz;

    pz = ((Dyadic)(*Z))->Wert;
    while (pz != 0L) {
      qz = pz;
      pz = pz->BlockNext;
      LRFreeBlock(qz);
    }
    pz=LRDummyBlock;

    for (;;){
    if (px == 0L) {
       pz->BlockNext=0L;
       break;
       }
    else
       {
       pz->BlockNext=LRNewBlock();
       pz=pz->BlockNext;
       memcpy(pz->BlockFeld,px->BlockFeld,BLOCKSIZE*sizeof(mp_limb_t));
       px = px->BlockNext;
       }
    }


    ((Dyadic)(*Z))->Wert         = LRDummyBlock->BlockNext;
    ((Dyadic)(*Z))->highexponent = ((Dyadic)(*X))->highexponent;
    ((Dyadic)(*Z))->negativ      = ((Dyadic)(*X))->negativ;
    ((Dyadic)(*Z))->anzahl       = ((Dyadic)(*X))->anzahl;
}


void Dyadic_LRTRUNC
(const Dyadic_Dyadic *X, Dyadic_Dyadic *Z)
{
  if (((Dyadic)(*X))->highexponent <= 0) {
    Dyadic_INTTOLR(0, Z);
  } else {
    LRGET(X, ((Dyadic)(*X))->highexponent, AKKUMULATOR1, ((Dyadic)(*X))->highexponent);
    LRSET(Z, ((Dyadic)(*X))->highexponent, AKKUMULATOR1, ((Dyadic)(*X))->highexponent, ((Dyadic)(*X))->negativ);
  }
}

void Dyadic_LRINV
(const Dyadic_Dyadic *X, Dyadic_Dyadic *Z, int N)
{
  int n, Anzahl, XAnz;

  if (N - 1 >= 0) {
    n = (N - 1) / Dyadic_Base_BITZAHL;
  } else {
    n = N / Dyadic_Base_BITZAHL;
    n--;
  }
  if (n > -((Dyadic)(*X))->highexponent) {
    Dyadic_INTTOLR(0, Z);
  } else {
    AKKUMULATOR1[0] = 1;
    Anzahl = 2 - ((Dyadic)(*X))->highexponent - n;
    if (Anzahl < ((Dyadic)(*X))->anzahl) {
      XAnz = Anzahl;
    } else {
      XAnz = ((Dyadic)(*X))->anzahl;
    }
    LRGET(X, ((Dyadic)(*X))->highexponent, AKKUMULATOR2, XAnz);
    Dyadic_Base_SysDiv(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, 1, XAnz, Anzahl, FALSE);
    LRSET(Z, 2 - ((Dyadic)(*X))->highexponent, AKKUMULATOR3, Anzahl, ((Dyadic)(*X))->negativ);
  }
}

void Dyadic_LRSQRT
(const Dyadic_Dyadic *X, Dyadic_Dyadic *Z, int N)
{
  int i, k;
  BOOLEAN LAST;
 
  i = Dyadic_LRSIGN(X);
  if (i < 0) {
    (*Dyadic_LRError)(4);
    Dyadic_INTTOLR(0, Z);
    return;
  }
  if (i == 0) {
    Dyadic_INTTOLR(0, Z);
    return;
  }
  i = Dyadic_LRSIZE(X);
  if (i < 2 * N) {
    Dyadic_INTTOLR(0, Z);
    return;
  }
  i = i / 2;
  Dyadic_INTTOLR(1, &A_LRSQRT);
  Dyadic_LRSHIFT(&A_LRSQRT, &A_LRSQRT, i);
  k = i;
  LAST = FALSE;
  for (;;) {
    k = 2 * k - i - 1;
    if (!LAST && k <= (N + i) / 2 - 2) {
      LAST = TRUE;
      k = (N + i) / 2 - 2;
    } else if (k < N) {
      k = N - 1;
    }
    Dyadic_LRDIV(X, &A_LRSQRT, &B_LRSQRT, k - 10);
    Dyadic_LRADD(&B_LRSQRT, &A_LRSQRT, &A_LRSQRT, k - 10);
    Dyadic_LRSHIFT(&A_LRSQRT, &A_LRSQRT, -1);
    if (k < N) {
      goto EXIT_4;
    }
  } EXIT_4:;
  Dyadic_LRCOPY(&A_LRSQRT, Z, N - 1);
}

void BEGIN_Dyadic()
{
  static BOOLEAN has_been_called = FALSE;

  if (!has_been_called) {
    has_been_called = TRUE;

    Initialize();
  }
}
