#include <string>
#include <stdlib.h>
#include <iostream>
#include <stdio.h>

#include "gmp.h"
//#include "gmp-impl.h"
#include "cGMP_ext.h"
#include "longlong.h"

#define DEBUG(x)

int ext_gmp_var_count=0;

#define MaxFreeVars 1000
ext_gmp_type gmp_FreeVars[MaxFreeVars];
int gmp_FreeVarCount=0L;


/****************** eigenes hashing *********************/

/* ---STRUKTUREN--- */
struct our_key_type {
  unsigned int p1, p2;
  unsigned char op;
};

struct listnode {
  unsigned long long p1;
  unsigned long long p2;
  unsigned char op;
  hash_record_pointer record_pointer;
  struct listnode *next;
};

struct INTEGER_listnode{
  unsigned long long INTEGER_id;
  int_gmp_type INTEGER_data;
  struct INTEGER_listnode *next;  
};

/* ---globale Variablen---*/
struct listnode *List_End;
struct listnode *MAP;
struct listnode *found;
unsigned int max_id=0; 
const unsigned int max_table_size=32768;

struct INTEGER_listnode *INTEGER_List_End;
struct INTEGER_listnode *INTEGER_MAP;
unsigned int max_INT_id=0;
const unsigned int max_INTEGER_table_size=16384;

unsigned int in_table, in_INTEGER_table;

/* ---PROTOTYPEN--- */
struct listnode *create_hashmap(int buckets);
static void cl_hashmap(struct listnode *map, int buckets);
static void insert(struct listnode *map, unsigned int mapsize, unsigned long long p1, unsigned long long p2, unsigned char op, hash_record *hr);
static hash_record inline *find(struct listnode *map, unsigned int mapsize, unsigned long long p1, unsigned long long p2, unsigned char op);
struct INTEGER_listnode inline *create_INTEGER_hashmap(unsigned int buckets);
static void inline cl_INTEGER_hashmap(struct INTEGER_listnode *map, unsigned int buckets);
static void inline INTEGER_insert(struct INTEGER_listnode *map, unsigned int mapsize, unsigned long long p1, int_gmp_type data);
static unsigned long long inline INTEGER_find(struct INTEGER_listnode *map, unsigned int mapsize, int_gmp_type data);

/* ---FUNKTIONEN--- */

/* legt neue Hashmap an */
struct listnode inline *create_hashmap(unsigned int buckets) {
  struct listnode *map;
  unsigned int i;

  map = (struct listnode *)malloc(2 * buckets * sizeof(struct listnode));
  for(i=0; i<2 * buckets; ++i) {
    map[i].p1 = 0;
    map[i].p2 = 0;
    map[i].op = 0;
    map[i].record_pointer = NULL;
    map[i].next = NULL;
  }
  List_End = &map[buckets-1];
  in_table = 0;
  return map;
}

static void inline cl_hashmap(struct listnode *map, unsigned int buckets) {
  unsigned int i;
  for(i=0; i<2 * buckets; ++i) {
    map[i].p1 = 0;
    map[i].p2 = 0;
    map[i].op = 0;
    map[i].next = NULL;
    rfc_clear(&(map[i].record_pointer));  
    map[i].record_pointer = NULL;
  }
  List_End = &map[buckets-1];
  in_table = 0;
  return;
}

/* fuegt Hashwert und Pointer in Hashmap ein (wenn Platz, dann
   direkt im Array, sonst an erster Stelle in der Bucket-Liste)*/
static void inline insert(struct listnode *map, unsigned int mapsize, unsigned long long p1, unsigned long long p2, unsigned char op, hash_record *hr) {
  struct listnode *lp, *dummy;
  unsigned int hash = ((p1<<16)+(p1>>16)+p2+(op<<24)) % mapsize;
  
  lp = &map[hash];
  if (lp->op == 0){
    map[hash].p1 = p1;
    map[hash].p2 = p2;
    map[hash].op = op;
    map[hash].record_pointer = hr;
  }  
  else{  
    if (found != NULL){
      found->p1 = p1;
      found->p2 = p2;
      found->op = op;
      found->record_pointer = hr;
    }
    else{
      dummy = ++List_End;
      dummy->p1 = p1;
      dummy->p2 = p2;
      dummy->op = op;
      dummy->record_pointer = hr;
      dummy->next = lp->next;
      lp->next = dummy;
    }
  }
  in_table++;
}

/* findet listenknoten */
static hash_record inline *find(struct listnode *map, unsigned int mapsize, unsigned long long p1, unsigned long long p2, unsigned char op) {
  struct listnode *lp;
  unsigned int hash = ((p1<<16)+(p1>>16)+p2+(op<<24)) % mapsize;
  lp = &map[hash];
  do {
    if( lp->p1 == p1 && lp->p2 == p2 && lp->op == op){
      found = lp;
      return lp->record_pointer;
    }
    lp = lp->next;
  } while( lp != NULL );
  found = NULL;
  return NULL;
}

/*--- Funktionen fuer INTEGER-Tabelle ---*/

struct INTEGER_listnode inline *create_INTEGER_hashmap(unsigned int buckets) {
  struct INTEGER_listnode *map;
  unsigned int i;

  map = (struct INTEGER_listnode *)malloc(2 * buckets * sizeof(struct INTEGER_listnode));
  for(i=0; i<2 * buckets; ++i) {
    map[i].INTEGER_id = 0ll;
    map[i].next = NULL;
    map[i].INTEGER_data = NULL;
  }
  INTEGER_List_End = &map[buckets-1];
  in_INTEGER_table = 0;
  return map;
}

static void inline cl_INTEGER_hashmap(struct INTEGER_listnode *map, unsigned int buckets) {
  unsigned int i;
  for(i=0; i<2 * buckets; ++i) {
    map[i].INTEGER_id = 0ll;
    map[i].next = NULL;
    if (map[i].INTEGER_data != NULL)    int_gmp_free(map[i].INTEGER_data);
    map[i].INTEGER_data = NULL;
  }
  INTEGER_List_End = &map[buckets-1];
  in_INTEGER_table = 0;
  return;
}

static void inline INTEGER_insert(struct INTEGER_listnode *map, unsigned int mapsize, unsigned long long p1, int_gmp_type data) {
  struct INTEGER_listnode *lp, *dummy;
  unsigned int hash = mpz_fdiv_ui(data, 2147483647) % mapsize;
  lp = &map[hash];
  if (lp->INTEGER_id == 0){
    map[hash].INTEGER_id = p1;
    map[hash].INTEGER_data = data;
  }  
  else{  

  dummy = ++INTEGER_List_End;
  dummy->INTEGER_id = p1;
  dummy->INTEGER_data = data;
  dummy->next = lp->next;
  map[hash].next = dummy;
  }
  in_INTEGER_table++;
}

static unsigned long long inline INTEGER_find(struct INTEGER_listnode *map, unsigned int mapsize, int_gmp_type data) {
  struct INTEGER_listnode *lp;
  lp = &map[mpz_fdiv_ui(data, 2147483647) % mapsize];
  do {
    if( lp->INTEGER_id != 0ll && mpz_cmp(lp->INTEGER_data,data) == 0)
      return lp->INTEGER_id;
    lp = lp->next;
  } while( lp != NULL );
  return 0ll;
}
/****************** eigenes hashing Ende*********************/

hash_record* save_record_list[SAVE_RECORD_LIST_SIZE ]; // als Stack verwaltetes Feld
unsigned int save_record_list_pointer = SAVE_RECORD_LIST_SIZE ; // zeigt auf das erste leere Element im Stack

inline void save_record_list_init(){
  for (unsigned int i = 0; i < SAVE_RECORD_LIST_SIZE ; i++){
    save_record_list[i] = new(hash_record);
    save_record_list[i]->data=ext_gmp_init();
  }
}

unsigned int hashtable_init(){
  ext_gmp_initialize();
  save_record_list_init();
  MAP = create_hashmap(max_table_size);
  INTEGER_MAP = create_INTEGER_hashmap(max_INTEGER_table_size);
  return 1;
};

unsigned int hashtable_init_marker=hashtable_init();

//Es folgen Funktionen mit eigener Hash-Tabellen-Implementierung

//Folgende Funktion arbeitet auf der Integer-Tabelle
void hash_INTEGER_to_mp(const int_gmp_type i ,hash_record_pointer * z) {

  hash_record_pointer temp_pointer = NULL;
  int_gmp_type j=int_gmp_init();
  mpz_set(j,i);
  
  if ( in_INTEGER_table > max_INTEGER_table_size ) cl_INTEGER_hashmap(INTEGER_MAP,max_INTEGER_table_size);
  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);
  unsigned long long p = INTEGER_find(INTEGER_MAP, max_INTEGER_table_size, j);  
  
  if (p == 0ll ){
    INTEGER_insert(INTEGER_MAP, max_INTEGER_table_size, ++max_INT_id, j);
    p = max_INT_id;
  }
  else{
    int_gmp_free(j);
  }
  temp_pointer = find(MAP, max_table_size, (unsigned long long) p, (unsigned long long)0, (unsigned char)'@');
  hash_record_pointer * r = &temp_pointer;
  if ((*r) != NULL) {
    rfc_clear(z);
    (*r)->rfc++;
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:found-I(%uld):",(*z)->id);)
  } 
  else { 
    (*r)=getrecord();
    ext_gmp_set_z((*r)->data,i);
    (*r)->id=++max_id;
    (*r)->rfc+=2;
    insert(MAP, max_table_size,(long long)  p, (long long)0, (unsigned char)'@', *r);     
    rfc_clear(z);
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:I(%uld):",(*z)->id);)
  }
} 

void hash_int_to_mp(int i,hash_record_pointer * z) {

  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);

  hash_record_pointer temp_pointer = find(MAP, max_table_size, i+4294967296ull, 0ull, (unsigned char)'#');
  hash_record_pointer * r = &temp_pointer;

  if ((*r) !=NULL ) {
    rfc_clear(z);
    (*r)->rfc++;
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:found-i(%uld)=%uld:",(*z)->id,i);)
  } 
  else {
    (*r)=getrecord();
    mpf_set_si((*r)->data,i);    
    (*r)->id=++max_id;
    (*r)->rfc+=2;
    insert(MAP, max_table_size, i+4294967296ull, 0ull, (unsigned char)'#', *r);    
    rfc_clear(z);
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:i(%uld)=%uld:",(*z)->id,i);)
  }
} 

/*NM*** double hat soviele Bits wie long long, daher vereinfacht*/
void hash_double_to_mp(double d,hash_record_pointer * z) {
  unsigned long long dll=*(reinterpret_cast<unsigned long long *>(&d));
      
  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);

  hash_record_pointer temp_pointer = find(MAP, max_table_size, dll, 0, (unsigned char)'d');
  hash_record_pointer * r = &temp_pointer;

  if ((*r) !=NULL ) {
    rfc_clear(z);
    (*r)->rfc++;
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:found-d(%uld)=%f:",(*z)->id,d);)
  } 
  else {
    (*r)=getrecord();
    mpf_set_d((*r)->data,d);    
    (*r)->id=++max_id;
    (*r)->rfc+=2;
    insert(MAP, max_table_size, dll, 0, (unsigned char)'d', *r);    
    rfc_clear(z);
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:i(%uld)=%f:",(*z)->id,d);)
  }
}

void hash_add  (const hash_record_pointer z1, 
                const hash_record_pointer z2,
		      hash_record_pointer* z,int p) {
  if (z1->id==0) {z1->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z1->id));}
  if (z2->id==0) {z2->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z2->id));}
 
  listnode temp_key;  
  if (z1->id < z2->id) { 
    temp_key.p1=z1->id;
    temp_key.p2=z2->id;
  }
  else { 
    temp_key.p1=z2->id;
    temp_key.p2=z1->id;
  }

  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);
  
  hash_record_pointer temp_pointer = find(MAP, max_table_size, temp_key.p1, temp_key.p2, (unsigned char)'+');
  if (temp_pointer !=NULL && temp_pointer->prec <= p) {
    rfc_clear(z);
    temp_pointer->rfc++;
    (*z)=temp_pointer;
    DEBUG(fprintf(stderr,"\n:found-+(%uld,%uld,%uld,%uld):",z1->id,z2->id,(*z)->id,(*z)->rfc);)
  } 
  else {
    rfc_clear(&temp_pointer);
    temp_pointer=getrecord();
    
    ext_gmp_add(z1->data,z2->data,temp_pointer->data,p);
    temp_pointer->id=++max_id;
    temp_pointer->rfc+=2;
    temp_pointer->prec=p;
    insert(MAP, max_table_size, temp_key.p1, temp_key.p2, (unsigned char)'+', temp_pointer);
    rfc_clear(z);
    (*z)=temp_pointer;
    DEBUG(fprintf(stderr,"\n:+(%uld,%uld,%uld,%uld):",z1->id,z2->id,(*z)->id,(*z)->rfc);)
  }
} 

void hash_mul  (const hash_record_pointer z1, 
                const hash_record_pointer z2,
		      hash_record_pointer* z,int p){
  // das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! 
  if (z1->id==0) {z1->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z1->id));}
  if (z2->id==0) {z2->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z2->id));}
 
  listnode temp_key;
  if (z1->id < z2->id) { 
    temp_key.p1=z1->id;
    temp_key.p2=z2->id;
  }
  else {
    temp_key.p1=z2->id;
    temp_key.p2=z1->id;
  }
      
  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);

  hash_record_pointer temp_pointer = find(MAP, max_table_size, temp_key.p1, temp_key.p2, (unsigned char)'*');
  hash_record_pointer * r = &temp_pointer;

  if ((*r) !=NULL && (*r)->prec <= p) {
    rfc_clear(z);
    (*r)->rfc++;
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:found-*(%uld,%uld,%uld,%uld):",z1->id,z2->id,(*z)->id,(*z)->rfc);)
  } 
  else {
    rfc_clear(r);
    (*r)=getrecord();
    
    ext_gmp_mul(z1->data,z2->data,(*r)->data,p);
    (*r)->id=++max_id;
    (*r)->rfc+=2;
    (*r)->prec=p;
    insert(MAP, max_table_size, temp_key.p1, temp_key.p2, (unsigned char)'*', *r);    
    rfc_clear(z);
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:*(%uld,%uld,%uld,%uld):",z1->id,z2->id,(*z)->id,(*z)->rfc);)
  }
} 

void hash_div  (const hash_record_pointer z1, 
                const hash_record_pointer z2,
		      hash_record_pointer* z,int p){
  // das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! 
  if (z1->id==0) {z1->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z1->id));}
  if (z2->id==0) {z2->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z2->id));}
 
  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);

  hash_record_pointer temp_pointer = find(MAP, max_table_size, z1->id, z2->id, (unsigned char)'/');
  hash_record_pointer * r = &temp_pointer;

  if ((*r) !=NULL && (*r)->prec <= p) {
    rfc_clear(z);
    (*r)->rfc++;
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:found-/(%uld,%uld,%uld,%uld):",z1->id,z2->id,(*z)->id,(*z)->rfc);)
  } 
  else {
    rfc_clear(r);
    (*r)=getrecord();
    
    ext_gmp_div(z1->data,z2->data,(*r)->data,p);
    (*r)->id=++max_id;
    (*r)->rfc+=2;
    (*r)->prec=p;
    insert(MAP, max_table_size, z1->id, z2->id, (unsigned char)'/', *r);    
    rfc_clear(z);
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:/(%uld,%uld,%uld):",z1->id,z2->id,(*z)->id);)
  }
} 

void hash_divi (const hash_record_pointer z1, 
                const int z2,
		      hash_record_pointer* z,int p){
  // das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! 
  if (z1->id==0) {z1->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z1->id));}
  
  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);

  hash_record_pointer temp_pointer = find(MAP, max_table_size, z1->id, z2+4294967296ull, (unsigned char)'\\');
  hash_record_pointer * r = &temp_pointer;

  if ((*r) !=NULL && (*r)->prec <= p) {
    rfc_clear(z);
    (*r)->rfc++;
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:found-divi(%uld):",z1->id);)
  } 
  else {
    rfc_clear(r);
    (*r)=getrecord();
    
    ext_gmp_div_i(z1->data,z2,(*r)->data,p);
    (*r)->id=++max_id;
    (*r)->rfc+=2;
    (*r)->prec=p;
    insert(MAP, max_table_size, z1->id, z2+4294967296ull, (unsigned char)'\\', *r);    
    rfc_clear(z);
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:divi(%uld,%uld,%uld):",z1->id,(*z)->id,(*z)->rfc);)
  }
} 

void hash_addi (const hash_record_pointer z1, 
                const int z2,
		      hash_record_pointer* z,int p){
  // das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! 
  if (z1->id==0) {z1->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z1->id));}
  
  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);

  hash_record_pointer temp_pointer = find(MAP, max_table_size, z1->id, z2+4294967296ull, (unsigned char)'a');
  hash_record_pointer * r = &temp_pointer;

  if ((*r) !=NULL && (*r)->prec <= p) {
    rfc_clear(z);
    (*r)->rfc++;
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:found-addi(%uld,%uld,%uld):",z1->id,(*z)->id,(*z)->rfc);)
  } 
  else {
    rfc_clear(r);
    (*r)=getrecord();
    
    ext_gmp_add_i(z1->data,z2,(*r)->data,p);
    (*r)->id=++max_id;
    (*r)->rfc+=2;
    (*r)->prec=p;
    insert(MAP, max_table_size, z1->id, z2+4294967296ull, (unsigned char)'a', *r);    
    rfc_clear(z);
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:addi(%uld,%uld,%uld):",z1->id,(*z)->id,(*z)->rfc);)
  }
} 

void hash_subi (const hash_record_pointer z1, 
                const int z2,
		      hash_record_pointer* z,int p){
  // das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! 
  if (z1->id==0) {z1->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z1->id));}
  
  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);

  hash_record_pointer temp_pointer = find(MAP, max_table_size, z1->id, z2+4294967296ull, (unsigned char)'b');
  hash_record_pointer * r = &temp_pointer;

  if ((*r) !=NULL && (*r)->prec <= p) {
    rfc_clear(z);
    (*r)->rfc++;
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:found-subi(%uld,%uld,%uld,%uld):",z1->id,z2,(*z)->id,(*z)->rfc);)
  } 
  else {
    rfc_clear(r);
    (*r)=getrecord();
    
    ext_gmp_sub_i(z1->data,z2,(*r)->data,p);
    (*r)->id=++max_id;
    (*r)->rfc+=2;
    (*r)->prec=p;
    insert(MAP, max_table_size, z1->id, z2+4294967296ull, (unsigned char)'b', *r);    
    rfc_clear(z);
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:subi(%uld,%uld,%uld,%uld):",z1->id,z2,(*z)->id,(*z)->rfc);)
  }
} 

void hash_muli (const hash_record_pointer z1, 
                const int z2,
		      hash_record_pointer* z,int p){
  // das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! 
  if (z1->id==0) {z1->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z1->id));}

  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);

  hash_record_pointer temp_pointer = find(MAP, max_table_size, z1->id, z2+4294967296ull, (unsigned char)'m');
  hash_record_pointer * r = &temp_pointer;

  if ((*r) !=NULL && (*r)->prec <= p) {
    rfc_clear(z);
    (*r)->rfc++;
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:found-muli(%uld,%uld,%uld,%uld):",z1->id,z2,(*z)->id,(*z)->rfc);)
  } 
  else {
    rfc_clear(r);
    (*r)=getrecord();
    
    ext_gmp_mul_i(z1->data,z2,(*r)->data,p);
    (*r)->id=++max_id;
    (*r)->rfc+=2;
    (*r)->prec=p;
    insert(MAP, max_table_size, z1->id, z2+4294967296ull, (unsigned char)'m', *r);    
    rfc_clear(z);
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:muli(%uld,%uld,%uld,%uld):",z1->id,z2,(*z)->id,(*z)->rfc);)
  }
} 

void hash_idiv (const int z1,
                const hash_record_pointer z2,
		      hash_record_pointer* z,int p){
  // das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! 
  if (z2->id==0) {z2->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z2->id));}
  
  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);

  hash_record_pointer temp_pointer = find(MAP, max_table_size, z2->id, z1+4294967296ull, (unsigned char)'(');
  hash_record_pointer * r = &temp_pointer;

  if ((*r) !=NULL && (*r)->prec <= p) {
    rfc_clear(z);
    (*r)->rfc++;
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:found-idiv(%uld,%uld,%uld,%uld):",z1,z2->id,(*z)->id,(*z)->rfc);)
  } 
  else {
    rfc_clear(r);
    (*r)=getrecord();
    
    ext_gmp_i_div(z1,z2->data,(*r)->data,p);
    (*r)->id=++max_id;
    (*r)->rfc+=2;
    (*r)->prec=p;
    insert(MAP, max_table_size, z2->id, z1+4294967296ull, (unsigned char)'(', *r);    
    rfc_clear(z);
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:idiv(%uld,%uld,%uld,%uld):",z1,z2->id,(*z)->id,(*z)->rfc);)
  }
} 

void hash_isub (const int z1,
                const hash_record_pointer z2,
		      hash_record_pointer* z,int p){
  // das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! 
  if (z2->id==0) {z2->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z2->id));}
  
  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);

  hash_record_pointer temp_pointer = find(MAP, max_table_size, z2->id, z1+4294967296ull, (unsigned char)')');
  hash_record_pointer * r = &temp_pointer;

  if ((*r) !=NULL && (*r)->prec <= p) {
    rfc_clear(z);
    (*r)->rfc++;
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:found-isub(%uld,%uld,%uld,%uld):",z1,z2->id,(*z)->id,(*z)->rfc);)
  } 
  else {
    rfc_clear(r);
    (*r)=getrecord();
    
    ext_gmp_i_sub(z1,z2->data,(*r)->data,p);
    (*r)->id=++max_id;
    (*r)->rfc+=2;
    (*r)->prec=p;
    insert(MAP, max_table_size, z2->id, z1+4294967296ull, (unsigned char)')', *r);    
    rfc_clear(z);
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:isub(%uld,%uld,%uld,%uld):",z1,z2->id,(*z)->id,(*z)->rfc);)
  }
} 

void hash_sub  (const hash_record_pointer z1, 
                const hash_record_pointer z2,
		      hash_record_pointer* z,int p){
  // das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! 
  if (z1->id==0) {z1->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z1->id));}
  if (z2->id==0) {z2->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z2->id));}
      
  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);

  hash_record_pointer temp_pointer = find(MAP, max_table_size, z1->id, z2->id, (unsigned char)'-');
  hash_record_pointer * r = &temp_pointer;

  if ((*r) !=NULL && (*r)->prec <= p) {
    rfc_clear(z);
    (*r)->rfc++;
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:found--(%uld,%uld,%uld,%uld):",z1->id,z2->id,(*z)->id,(*z)->rfc);)
  } 
  else {
    rfc_clear(r);
    (*r)=getrecord();
    
    ext_gmp_sub(z1->data,z2->data,(*r)->data,p);
    (*r)->id=++max_id;
    (*r)->rfc+=2;
    (*r)->prec=p;
    insert(MAP, max_table_size, z1->id, z2->id, (unsigned char)'-', *r);    
    rfc_clear(z);
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:-(%uld,%uld,%uld):",z1->id,z2->id,(*z)->id);)
  }
} 

void hash_sqrt (const hash_record_pointer z1, 
		      hash_record_pointer* z,int p){
  // das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! 
  if (z1->id==0) {z1->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z1->id));}

  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);

  hash_record_pointer temp_pointer = find(MAP, max_table_size, z1->id, 0, (unsigned char)'s');
  hash_record_pointer * r = &temp_pointer;

  if ((*r) !=NULL && (*r)->prec <= p) {
    rfc_clear(z);
    (*r)->rfc++;
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:found-sqrt(%uld,%uld,%uld,%uld):",z1->id,(*z)->id,(*z)->rfc);)
  } 
  else {
    rfc_clear(r);
    (*r)=getrecord();
    
    ext_gmp_sqrt(z1->data,(*r)->data,p);
    (*r)->id=++max_id;
    (*r)->rfc+=2;
    (*r)->prec=p;
    insert(MAP, max_table_size, z1->id, 0, (unsigned char)'s', *r);    
    rfc_clear(z);
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:sqrt(%uld,%uld,%uld):",z1->id,(*z)->id,(*z)->rfc);)
  }
} 

void hash_shift (const hash_record_pointer z1, 
		      hash_record_pointer* z,int sw){
  // das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! 
  if (z1->id==0) {z1->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z1->id));}

  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);

  hash_record_pointer temp_pointer = find(MAP, max_table_size, z1->id, sw+4294967296ull, (unsigned char)'S');
  hash_record_pointer * r = &temp_pointer;

  if ((*r) !=NULL) {
    rfc_clear(z);
    (*r)->rfc++;
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:found-shift(%uld,%uld,%uld):",z1->id,(*z)->id,(*z)->rfc);)
  } 
  else {
    rfc_clear(r);
    (*r)=getrecord();
    
    ext_gmp_shift(z1->data,(*r)->data,sw);
    (*r)->id=++max_id;
    (*r)->rfc+=2;
    insert(MAP, max_table_size, z1->id, sw+4294967296ull, (unsigned char)'S', *r);    
    rfc_clear(z);
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:shift(%uld,%uld,%uld):",z1->id,(*z)->id,(*z)->id);)
  }
} 

void hash_abs  (const hash_record_pointer z1, 
		      hash_record_pointer* z){
  // das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! 
  if (z1->id==0) {z1->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z1->id));}

  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);

  hash_record_pointer temp_pointer = find(MAP, max_table_size, z1->id, 0, (unsigned char)'c');
  hash_record_pointer * r = &temp_pointer;

  if ((*r) !=NULL) {
    rfc_clear(z);
    (*r)->rfc++;
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:found-abs(%uld,%uld,%uld):",z1->id,(*z)->id,(*z)->rfc);)
  } 
  else {
    rfc_clear(r);
    (*r)=getrecord();
    
    ext_gmp_abs(z1->data,(*r)->data);
    (*r)->id=++max_id;
    (*r)->rfc+=2;
    insert(MAP, max_table_size, z1->id, 0, (unsigned char)'c', *r);    
    rfc_clear(z);
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:abs(%uld,%uld,%uld):",z1->id,(*z)->id,(*z)->id);)
  }
} 

void hash_truncate (const hash_record_pointer z1, 
		          hash_record_pointer* z){
  // das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! 
  if (z1->id==0) {z1->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z1->id));}
      
  if ( in_table > max_table_size ) cl_hashmap(MAP,max_table_size);

  hash_record_pointer temp_pointer = find(MAP, max_table_size, z1->id, 0, (unsigned char)'t');
  hash_record_pointer * r = &temp_pointer;

  if ((*r) !=NULL) {
    rfc_clear(z);
    (*r)->rfc++;
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:found-t(%uld,%uld,%uld):",z1->id,(*z)->id,(*z)->rfc);)
  } 
  else {
    rfc_clear(r);
    (*r)=getrecord();
    
    ext_gmp_truncate(z1->data,(*r)->data);
    (*r)->id=++max_id;
    (*r)->rfc+=2;
    insert(MAP, max_table_size, z1->id, 0, (unsigned char)'t', *r);    
    rfc_clear(z);
    (*z)=(*r);
    DEBUG(fprintf(stderr,"\n:trunc(%uld,%uld,%uld):",z1->id,(*z)->id,(*z)->id);)
  }
} 

/***************************************************************************************/
/***************************************************************************************/
/***************************************************************************************/
#define DEFINE_BLOCK(r) hash_record_pointer r=getrecord()
#define SEGFAULT ((hash_record_pointer)NULL)->id;

inline void COPY_BLOCK(hash_record_pointer r,hash_record_pointer* z,int p){
    r->id=++max_id;
    r->rfc++;
    r->prec=p;
    rfc_clear(z);
    (*z)=r;
}

void hash_sub_det  (const hash_record_pointer z1, 
                const hash_record_pointer z2,
		      hash_record_pointer* z,int p){
  /* das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! */
  if (z1->id==0) {z1->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z1->id));}
  if (z2->id==0) {z2->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z2->id));}

    DEFINE_BLOCK(r);
    ext_gmp_sub(z1->data,z2->data,r->data,p);
    COPY_BLOCK(r,z,p);
    DEBUG(fprintf(stderr,"\n:*(%uld,%uld,%uld,%uld):",z1->id,z2->id,(*z)->id,(*z)->rfc);)
} 

void hash_add_det  (const hash_record_pointer z1, 
                const hash_record_pointer z2,
		      hash_record_pointer* z,int p){
  /* das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! */
  if (z1->id==0) {z1->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z1->id));}
  if (z2->id==0) {z2->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z2->id));}

    DEFINE_BLOCK(r);
    ext_gmp_add(z1->data,z2->data,r->data,p);
    COPY_BLOCK(r,z,p);
    DEBUG(fprintf(stderr,"\n:*(%uld,%uld,%uld,%uld):",z1->id,z2->id,(*z)->id,(*z)->rfc);)
} 

void hash_mul_det  (const hash_record_pointer z1, 
                const hash_record_pointer z2,
		      hash_record_pointer* z,int p){
  /* das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! */
  if (z1->id==0) {z1->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z1->id));}
  if (z2->id==0) {z2->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z2->id));}

    DEFINE_BLOCK(r);
    ext_gmp_mul(z1->data,z2->data,r->data,p);
    COPY_BLOCK(r,z,p);
    DEBUG(fprintf(stderr,"\n:*(%uld,%uld,%uld,%uld):",z1->id,z2->id,(*z)->id,(*z)->rfc);)
} 

void hash_div_det  (const hash_record_pointer z1, 
                const hash_record_pointer z2,
		      hash_record_pointer* z,int p){
  /* das folgende sollte nicht mehr vorkommen, wenn alle Konstruktoren umgesetzt sind! */
  if (z1->id==0) {z1->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z1->id));}
  if (z2->id==0) {z2->id=++max_id;DEBUG(fprintf(stderr,":%uld:",z2->id));}

    DEFINE_BLOCK(r);
    ext_gmp_div(z1->data,z2->data,r->data,p);
    COPY_BLOCK(r,z,p);
    DEBUG(fprintf(stderr,"\n:*(%uld,%uld,%uld,%uld):",z1->id,z2->id,(*z)->id,(*z)->rfc);)
} 

void hash_addi_det  (const hash_record_pointer z1, const int z2, hash_record_pointer* z, int p){SEGFAULT};
void hash_muli_det  (const hash_record_pointer z1, const int z2, hash_record_pointer* z, int p){SEGFAULT};
void hash_divi_det  (const hash_record_pointer z1, const int z2, hash_record_pointer* z, int p){SEGFAULT};
void hash_isub_det  (const int z1, const hash_record_pointer z2, hash_record_pointer* z, int p){SEGFAULT};
void hash_idiv_det  (const int z1, const hash_record_pointer z2, hash_record_pointer* z, int p){SEGFAULT};

/****************************************************************************/


