/*

STREAMS.cc -- implememtation of the C++-style IO for the iRRAM library
 
Copyright (C) 2004 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/


#include "iRRAM_core.h"

#undef cout
#undef cin
#undef ofstream
#undef ifstream

#include <fstream>

orstream rout;
irstream rin;


long long rstream::requests = 0;
long long rstream::outputs = 0;
unsigned int orstream::real_w=20;
unsigned int orstream::real_f =iRRAM_float_absolute;
bool inReiterate=false;

orstream::orstream(){ target = &cout;}
irstream::irstream(){target = &cin;}


void orstream::rewind(){ requests = 0; }
void orstream::reset() { requests = 0; outputs = 0; }

#define iRRAM_outexec(x) \
  if (inReiterate) {\
    DEBUG1(0,"ignored I/O-deletion in iteration!\n");\
    return;\
  }\
  if (ACTUAL_STACK.inlimit>0) {\
    DEBUG1(0,"illegal output in continuous section!\n");\
    return;\
  }\
  if (ACTUAL_STACK.inlimit==0){\
  ACTUAL_STACK.inlimit++;\
  if ((++requests) > outputs){\
    x;\
    outputs++;\
  }\
  ACTUAL_STACK.inlimit--;}else{ x;}


#define iRRAM_out(x) \
  if (ACTUAL_STACK.inlimit>0) {\
    DEBUG1(0,"illegal output in continuous section!\n");\
    return *this;\
  }\
  if (ACTUAL_STACK.inlimit==0){\
  ACTUAL_STACK.inlimit++;\
  if ((++requests) > outputs){\
    *target << x;\
    outputs++;\
  }\
  ACTUAL_STACK.inlimit--;}else{*target << x;} \
  return *this;
  
#define iRRAM_out2(x) \
  if (ACTUAL_STACK.inlimit>0) {\
    DEBUG1(0,"illegal output in continuous section!\n");\
    return *this;\
  }\
  if (ACTUAL_STACK.inlimit==0){\
  ACTUAL_STACK.inlimit++;\
  if ((++requests) > outputs){\
    x;\
    outputs++;\
  }\
  ACTUAL_STACK.inlimit--;}else{x;} \
  return *this;
  
orstream& orstream::operator<<(const bool c)  {iRRAM_out(c);}
orstream& orstream::operator<<(const char* c)  {iRRAM_out(c);}
orstream& orstream::operator<<(const short i)    {iRRAM_out(i);}
orstream& orstream::operator<<(const unsigned short i)    {iRRAM_out(i);}
orstream& orstream::operator<<(const int i)    {iRRAM_out(i);}
orstream& orstream::operator<<(const unsigned int ui)    {iRRAM_out(ui);}
orstream& orstream::operator<<(const long i)    {iRRAM_out(i);}
orstream& orstream::operator<<(const unsigned long ui)    {iRRAM_out(ui);}
orstream& orstream::operator<<(const double d) {iRRAM_out(d);}
orstream& orstream::operator<<(const float d) {iRRAM_out(d);}
orstream& orstream::operator<<(const string& s) {iRRAM_out(s);}
orstream& orstream::operator<<(const long long ll){iRRAM_out(ll);}
orstream& orstream::operator<<(const unsigned long long ull){iRRAM_out(ull);}

orstream& orstream::operator<<(const REAL& x) {iRRAM_out(swrite(x,real_w,real_f));}
orstream& orstream::operator<<(const DYADIC& x) {iRRAM_out(swrite(x,real_w));}
orstream& orstream::operator<<(const INTEGER& x) {iRRAM_out(swrite(x,real_w));}
orstream& orstream::operator<<(const RATIONAL& x) {iRRAM_out(swrite(x,real_w));}
orstream& orstream::operator<<( _Setw _f) { iRRAM_out(_f);}
orstream& orstream::operator<<( _Setprecision _f) { iRRAM_out(_f);}
orstream& orstream::operator<<( _Setfill<char> _f) { iRRAM_out(_f);}
orstream& orstream::operator<<( _Setiosflags _f) { iRRAM_out(_f);}
orstream& orstream::operator<<(_Resetiosflags _f) { iRRAM_out(_f);}
orstream& orstream::operator<<(_Setbase _f) { iRRAM_out(_f);}
orstream& orstream::operator<<(ostream& _f (ostream&) ) { iRRAM_out(_f);}

orstream& orstream::operator<<( _SetRwidth _f) { iRRAM_out2(real_w=_f._M_n);}
orstream& orstream::operator<<( _SetRflags _f) { iRRAM_out2(real_f=_f._M_n);}

orstream::~orstream()   {if (target != &cout){ iRRAM_outexec(delete target);}}
irstream::~irstream()   {if (target != &cin ){ iRRAM_outexec(delete target);}}


#define iRRAM_in(VAR,CACHE) \
  if (ACTUAL_STACK.inlimit>0) {				\
    DEBUG1(0,"illegal input in continuous section!\n");	\
    return *this; \
  }							\
  if (ACTUAL_STACK.inlimit==0){\
  if (CACHE.get(VAR)){ return *this;}			\
  ACTUAL_STACK.inlimit++;				\
  *target >> VAR;						\
  ACTUAL_STACK.inlimit--;				\
  CACHE.put(VAR);}else{*target >> VAR;} \
  return *this;


irstream& irstream::operator>>(bool& b) {iRRAM_in(b,cache_b);}
irstream& irstream::operator>>(short& i) {iRRAM_in(i,cache_sh);}
irstream& irstream::operator>>(unsigned short& ui) {iRRAM_in(ui,cache_ush);}
irstream& irstream::operator>>(int& i) {iRRAM_in(i,cache_i);}
irstream& irstream::operator>>(unsigned int& ui) {iRRAM_in(ui,cache_ui);}
irstream& irstream::operator>>(long& l) {iRRAM_in(l,cache_l);}
irstream& irstream::operator>>(unsigned long& ul) {iRRAM_in(ul,cache_ul);}
irstream& irstream::operator>>(float& d) {iRRAM_in(d,cache_f);}
irstream& irstream::operator>>(double& d) {iRRAM_in(d,cache_d);}
irstream& irstream::operator>>(long long& ll) {iRRAM_in(ll,cache_ll);}
irstream& irstream::operator>>(unsigned long long& ull) {iRRAM_in(ull,cache_ull);}
irstream& irstream::operator>>(string& s) {iRRAM_in(s,cache_s);}

#define iRRAM_in2(VAR,DATA) 				\
  string s; 						\
  if (ACTUAL_STACK.inlimit>0) {				\
    DEBUG1(0,"illegal input in continuous section!\n");	\
    return *this;					\
  }							\
  if (ACTUAL_STACK.inlimit==0){				\
  if (cache_s.get(s)){VAR=DATA(s); return *this;}	\
  ACTUAL_STACK.inlimit++;				\
  *target >> s ;					\
  VAR=DATA(s);						\
  ACTUAL_STACK.inlimit--;				\
  cache_s.put(s);}else{*target >> s ;VAR=DATA(s);};			\
  return *this;

irstream&  irstream::operator>>(REAL& d){iRRAM_in2(d,REAL);}
irstream&  irstream::operator>>(INTEGER& d){iRRAM_in2(d,INTEGER);}
irstream&  irstream::operator>>(DYADIC& d){iRRAM_in2(d,DYADIC);}

orstream::orstream(string s, _Ios_Openmode mod){
  if (ACTUAL_STACK.inlimit>0) {				
    DEBUG1(0,"illegal input in continuous section!\n");	
    return;
  }
  if (ACTUAL_STACK.inlimit==0){\
  if (cache_os.get(target)){ return;}
  target = new ofstream(s.c_str(),mod);	
  cache_os.put(target);}else{target = new ofstream(s.c_str(),mod);} \
  return;
}

irstream::irstream(string s, _Ios_Openmode mod){
  if (ACTUAL_STACK.inlimit>0) {				
    DEBUG1(0,"illegal input in continuous section!\n");	
    return;
  }
  if (ACTUAL_STACK.inlimit==0){\
  if (cache_is.get(target)){ return;}
  target = new ifstream(s.c_str(),mod);	
  cache_is.put(target);}else{target = new ifstream(s.c_str(),mod);} \
  return;
}

#define iRRAM_inexec(VAR,CACHE,STMNT) \
  if (ACTUAL_STACK.inlimit>0) {				\
    DEBUG1(0,"illegal input in continuous section!\n");	\
    return VAR; \
  }							\
  if (ACTUAL_STACK.inlimit==0){\
  if (CACHE.get(VAR)){return VAR;}			\
  ACTUAL_STACK.inlimit++;				\
  STMNT;						\
  ACTUAL_STACK.inlimit--;				\
  CACHE.put(VAR);}else{STMNT;}\
  return VAR;

bool orstream::eof(){
bool test=false; iRRAM_inexec(test,cache_b,test=target->eof()) }

bool orstream::fail(){
bool test=false; iRRAM_inexec(test,cache_b,test=target->fail()) }

bool orstream::good(){
bool test=false; iRRAM_inexec(test,cache_b,test=target->good()) }

bool orstream::bad(){
bool test=false; iRRAM_inexec(test,cache_b,test=target->bad()) }

bool operator !(orstream& x){
bool test=false; iRRAM_inexec(test,cache_b,test=(x.target->fail())) }

orstream::operator bool(){
bool test=false; iRRAM_inexec(test,cache_b,test=!(target->fail())) }

bool irstream::eof(){
bool test=false; iRRAM_inexec(test,cache_b,test=target->eof()) }

bool irstream::fail(){
bool test=false; iRRAM_inexec(test,cache_b,test=target->fail()) }

bool irstream::good(){
bool test=false; iRRAM_inexec(test,cache_b,test=target->good()) }

bool irstream::bad(){
bool test=false; iRRAM_inexec(test,cache_b,test=target->bad()) }

bool operator !(irstream& x){
bool test=false; iRRAM_inexec(test,cache_b,test=(x.target->fail())) }

irstream::operator bool(){
bool test=false; iRRAM_inexec(test,cache_b,test=!(target->fail())) }



